"""FileWatcher based on watchdog. Looks for changes in files with .pic.yml extension."""

from __future__ import annotations

import logging
import pathlib
import sys
import threading
import time
import traceback
from collections.abc import Callable

from IPython.terminal.embed import embed
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from gdsfactory.config import cwd
from gdsfactory.pdk import get_active_pdk, on_pdk_activated
from gdsfactory.read.from_yaml_template import cell_from_yaml_template
from gdsfactory.typings import ComponentSpec, PathType


class FileWatcher(FileSystemEventHandler):
    """Captures *.py or *.pic.yml file change events."""

    def __init__(self, logger=None, path: str | None = None) -> None:
        """Initialize the YAML event handler."""
        super().__init__()

        self.logger = logger or logging.root
        pdk = get_active_pdk()
        pdk.register_cells_yaml(dirpath=path, update=True)

        self.observer = Observer()
        self.path = path
        self.stopping = threading.Event()
        # if a new pdk happens to get activated during the watcher loop, reset accordingly
        on_pdk_activated.add_handler(self._on_pdk_activated)

    def _on_pdk_activated(self, new_pdk, old_pdk):
        from gdsfactory.cell import CACHE

        CACHE.clear()
        new_pdk.register_cells_yaml(dirpath=self.path, update=True)

    def start(self) -> None:
        self.observer.schedule(self, self.path, recursive=True)
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()

    def run(self) -> None:
        while not self.stopping.is_set():
            if not self.observer.is_alive():
                self.observer.start()
            time.sleep(1)
        self.observer.stop()
        self.observer.join()

    def stop(self) -> None:
        self.stopping.set()
        self.thread.join()

    def update_cell(self, src_path, update: bool = False) -> Callable:
        """Parses a YAML file to a cell function and registers into active pdk.

        Args:
            src_path: the path to the file
            update: if True, will update an existing cell function of the same name without raising an error
        Returns:
            The cell function parsed from the yaml file.

        """
        from gdsfactory.cell import CACHE

        pdk = get_active_pdk()
        print(f"Active PDK: {pdk.name}")
        filepath = pathlib.Path(src_path)
        cell_name = filepath.stem.split(".")[0]
        if cell_name in CACHE:
            CACHE.pop(cell_name)
        function = cell_from_yaml_template(filepath, name=cell_name)
        try:
            pdk.register_cells_yaml(**{cell_name: function}, update=update)
        except ValueError as e:
            print(e)
        return function

    def on_moved(self, event) -> None:
        super().on_moved(event)

        what = "directory" if event.is_directory else "file"
        if what == "file" and event.dest_path.endswith(".pic.yml"):
            self.logger.info("Moved %s: %s", what, event.src_path)
            self.update_cell(event.dest_path)
            self.get_component(event.src_path)

    def on_created(self, event) -> None:
        super().on_created(event)

        what = "directory" if event.is_directory else "file"
        if (
            what == "file"
            and event.src_path.endswith(".pic.yml")
            or event.src_path.endswith(".py")
        ):
            self.logger.info("Created %s: %s", what, event.src_path)
            self.get_component(event.src_path)

    def on_deleted(self, event) -> None:
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"

        if what == "file" and event.src_path.endswith(".pic.yml"):
            self.logger.info("Deleted %s: %s", what, event.src_path)
            pdk = get_active_pdk()
            filepath = pathlib.Path(event.src_path)
            cell_name = filepath.stem.split(".")[0]
            pdk.remove_cell(cell_name)

    def on_modified(self, event) -> None:
        super().on_modified(event)

        what = "directory" if event.is_directory else "file"
        if (
            what == "file"
            and event.src_path.endswith(".pic.yml")
            or event.src_path.endswith(".py")
        ):
            self.logger.info("Modified %s: %s", what, event.src_path)
            self.get_component(event.src_path)

    def update(self):
        pass

    def get_component(self, filepath):
        self.update()
        try:
            filepath = pathlib.Path(filepath)
            if filepath.exists():
                if str(filepath).endswith(".pic.yml"):
                    cell_func = self.update_cell(filepath, update=True)
                    c = cell_func()
                    c.show(show_ports=True)
                    # on_yaml_cell_modified.fire(c)
                    return c
                elif str(filepath).endswith(".py"):
                    d = dict(locals(), **globals())
                    d.update(__name__="__main__")
                    exec(filepath.read_text(), d, d)
                else:
                    print("Changed file {filepath} ignored (not .pic.yml or .py)")

        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            print(e)


def watch(path: PathType | None = cwd, pdk: str | None = None) -> None:
    path = str(path)
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    if pdk:
        get_active_pdk(name=pdk)
    watcher = FileWatcher(path=path)
    watcher.start()
    logging.info(
        f"File watcher looking for changes in *.py and *.pic.yml files in {path!r}. Stop with Ctrl+C"
    )
    embed()
    watcher.stop()


def show(component: ComponentSpec):
    import gdsfactory as gf

    c = gf.get_component(component)
    c.show()


if __name__ == "__main__":
    path = sys.argv[1] if len(sys.argv) > 1 else "."
    watch(path)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.geometry.boolean import boolean
from gdsfactory.geometry.offset import offset


@gf.cell
def outline(
    elements,
    distance=1,
    precision: float = 1e-3,
    join: str = "miter",
    tolerance: int = 2,
    join_first: bool = True,
    open_ports: bool | float = False,
    layer=0,
) -> Component:
    """Returns Component containing the outlined polygon(s).

    Creates an outline around all the polygons passed in the `elements`
    argument. `elements` may be a Component, Polygon, or list of Components.

    Args:
        elements: Component(/Reference), list of Component(/Reference), or Polygon
            Polygons to outline or Component containing polygons to outline.
        distance: int or float
            Distance to offset polygons. Positive values expand, negative shrink.
        precision: float
            Desired precision for rounding vertex coordinates.
        join: {'miter', 'bevel', 'round'}
            Type of join used to create the offset polygon.
        tolerance: int or float
            For miter joints, this number must be at least 2 and it represents the
            maximal distance in multiples of offset between new vertices and their
            original position before beveling to avoid spikes at acute joints. For
            round joints, it indicates the curvature resolution in number of
            points per full circle.
        open_ports: bool or float
            If not False, holes will be cut in the outline such that the Ports are
            not covered. If True, the holes will have the same width as the Ports.
            If a float, the holes will be be widened by that value (useful for fully
            clearing the outline around the Ports for positive-tone processes
        layer: int, array-like[2], or set
            Specific layer(s) to put polygon geometry on.).

    """
    layer = gf.get_layer(layer)
    gds_layer, gds_datatype = layer

    D = Component()
    if not isinstance(elements, list):
        elements = [elements]
    port_list = []
    for e in elements:
        if isinstance(e, Component):
            D.add_ref(e)
            port_list += list(e.ports.values())
        else:
            D.add(e)

    D_bloated = offset(
        D,
        distance=distance,
        join=join,
        precision=precision,
        tolerance=tolerance,
        layer=layer,
    )

    Trim = Component()
    if open_ports:
        trim_width = 0 if open_ports else open_ports * 2
        for port in port_list:
            trim = compass(size=(distance + 6 * precision, port.width + trim_width))
            trim_ref = Trim << trim
            trim_ref.connect("e3", port, overlap=2 * precision)

    Outline = boolean(
        A=D_bloated,
        B=[D, Trim],
        operation="A-B" if distance > 0 else "B-A",
        precision=precision,
        layer=layer,
    )
    if open_ports and len(elements) == 1:
        for port in port_list:
            Outline.add_port(port=port)
    return Outline


def test_outline() -> None:
    comp1 = gf.components.taper(length=100, width1=20, width2=50)
    c = gf.geometry.outline(comp1, open_ports=True, precision=1e-3)
    assert int(c.area()) == 206, int(c.area())


def test_outline_ports() -> None:
    e1 = gf.components.ellipse(radii=(6, 6))
    e2 = gf.components.ellipse(radii=(10, 4))
    c = outline([e1, e2])
    assert int(c.area()) == 52, int(c.area())


if __name__ == "__main__":
    # e1 = gf.components.ellipse(radii=(6, 6))
    # e2 = gf.components.ellipse(radii=(10, 4))
    # c = outline([e1, e2], distance=1)
    # c.show(show_ports=True)
    test_outline()
    test_outline_ports()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, Floats, LayerSpec

_gaps = (0.2,) * 10
_widths = (0.5,) * 10


@gf.cell
def grating_coupler_rectangular_arbitrary(
    gaps: Floats = _gaps,
    widths: Floats = _widths,
    width_grating: float = 11.0,
    length_taper: float = 150.0,
    polarization: str = "te",
    wavelength: float = 1.55,
    taper: ComponentSpec | None = taper_function,
    layer_grating: LayerSpec | None = None,
    layer_slab: LayerSpec = "SLAB150",
    slab_xmin: float = -1.0,
    slab_offset: float = 1.0,
    fiber_angle: float = 15,
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    r"""Grating coupler uniform with rectangular shape (not elliptical).
    Therefore it needs a longer taper.
    Grating teeth are straight instead of elliptical.

    Args:
        gaps: list of gaps between grating teeth.
        widths: list of grating widths.
        width_grating: grating teeth width.
        length_taper: taper length (um).
        polarization: 'te' or 'tm'.
        wavelength: in um.
        taper: function.
        layer_grating: Optional layer for grating. \
                by default None uses cross_section.layer. \
                if different from cross_section.layer expands taper.
        layer_slab: layer that protects the slab under the grating.
        slab_xmin: where 0 is at the start of the taper.
        slab_offset: from edge of grating to edge of the slab.
        fiber_angle: in degrees.
        cross_section: for input waveguide port.
        kwargs: cross_section settings.

    .. code::

                      fiber

                   /  /  /  /
                  /  /  /  /

                _|-|_|-|_|-|___ layer
                   layer_slab |
            o1  ______________|



        top view     _________
                    /| | | | |
                   / | | | | |
                  /taper_angle
                 /_ _| | | | |
        wg_width |   | | | | |
                 \   | | | | |
                  \  | | | | |
                   \ | | | | |
                    \|_|_|_|_|
                 <-->
                taper_length

    """
    xs = gf.get_cross_section(cross_section, **kwargs)
    wg_width = xs.width
    layer_wg = gf.get_layer(xs.layer)
    layer_grating = gf.get_layer(layer_grating) or layer_wg
    c = Component()

    if taper:
        taper_ref = c << gf.get_component(
            taper,
            length=length_taper,
            width2=width_grating,
            width1=wg_width,
            cross_section=xs,
        )

        c.add_port(port=taper_ref.ports["o1"], name="o1")
        xi = taper_ref.xmax
    else:
        length_taper = 0
        xi = 0

    widths = gf.snap.snap_to_grid(widths)
    gaps = gf.snap.snap_to_grid(gaps)

    y0 = width_grating / 2

    for width, gap in zip(widths, gaps):
        xi += gap
        points = snap_to_grid(
            np.array(
                [
                    [xi, -y0],
                    [xi, +y0],
                    [xi + width, +y0],
                    [xi + width, -y0],
                ]
            )
        )
        c.add_polygon(
            points,
            layer_grating,
        )
        xi += width

    if layer_slab:
        slab_xmin = length_taper - slab_offset
        slab_xmax = length_taper + np.sum(widths) + np.sum(gaps) + slab_offset
        slab_ysize = c.ysize + 2 * slab_offset
        yslab = slab_ysize / 2
        c.add_polygon(
            [
                (slab_xmin, yslab),
                (slab_xmax, yslab),
                (slab_xmax, -yslab),
                (slab_xmin, -yslab),
            ],
            layer_slab,
        )
    xport = np.round((xi + length_taper) / 2, 3)
    name = f"opt_{polarization.lower()}_{int(wavelength*1e3)}_{int(fiber_angle)}"
    c.add_port(
        name=name,
        port_type=name,
        center=(xport, 0),
        orientation=0,
        width=width_grating,
        layer=xs.layer,
    )
    c.info["polarization"] = polarization
    c.info["wavelength"] = wavelength

    gf.asserts.grating_coupler(c)
    if xs.add_bbox:
        c = xs.add_bbox(c)
    if xs.add_pins:
        c = xs.add_pins(c)
    return c


if __name__ == "__main__":
    c = grating_coupler_rectangular_arbitrary(width=3)
    # c = grating_coupler_rectangular_arbitrary(
    #     layer_grating=(3, 0), layer_slab=(2, 0), slab_offset=1
    # )
    c.show(show_ports=True)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Literal

import gdstk
from pydantic import BaseModel, Field

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component

if TYPE_CHECKING:
    from gdsfactory.technology import LayerViews


class LayerLevel(BaseModel):
    """Level for 3D LayerStack.

    Parameters:
        layer: (GDSII Layer number, GDSII datatype).
        thickness: layer thickness in um.
        thickness_tolerance: layer thickness tolerance in um.
        zmin: height position where material starts in um.
        zmin_tolerance: layer height tolerance in um.
        material: material name.
        sidewall_angle: in degrees with respect to normal.
        sidewall_angle_tolerance: in degrees.
        width_to_z: if sidewall_angle, relative z-position (0 --> zmin, 1 --> zmin + thickness).
        z_to_bias: parametrizes shrinking/expansion of the design GDS layer \
                when extruding from zmin (0) to zmin + thickness (1).\
                Defaults no buffering [[0, 1], [0, 0]].
        mesh_order: lower mesh order (1) will have priority over higher \
                mesh order (2) in the regions where materials overlap.
        refractive_index: int, complex or function that depends on wavelength (um).
        type: grow, etch, implant, or background.
        mode: octagon, taper, round. https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
        into: etch into another layer. https://gdsfactory.github.io/klayout_pyxs/DocGrow.html
        resistivity: for metals.
        bias: in um for the etch. Can be a single number or 2 numbers (bias_x, bias_y)
        derived_layer: Optional derived layer, used for layer_type='etch' to define the slab.
        info: simulation_info and other types of metadata.
        background_doping_concentration: uniform base doping level in the material (cm-3)
        background_doping_ion: uniform base doping ion in the material
        orientation: of the wafer (Miller indices of the plane)
    """

    name: str | None = None
    layer: tuple[int, int] | None = None
    thickness: float
    thickness_tolerance: float | None = None
    zmin: float
    zmin_tolerance: float | None = None
    material: str | None = None
    sidewall_angle: float = 0.0
    sidewall_angle_tolerance: float | None = None
    width_to_z: float = 0.0
    z_to_bias: tuple[list[float], list[float]] | None = None
    mesh_order: int = 3
    layer_type: Literal["grow", "etch", "implant", "background"] = "grow"
    mode: Literal["octagon", "taper", "round"] | None = None
    into: list[str] | None = None
    resistivity: float | None = None
    bias: tuple[float, float] | float | None = None
    derived_layer: tuple[int, int] | None = None
    info: dict[str, Any] = Field(default_factory=dict)
    background_doping_concentration: float | None = None
    background_doping_ion: str | None = None
    orientation: str | None = "100"

    @property
    def bounds(self) -> tuple[float, float]:
        """Calculates and returns the bounds of the layer level in the z-direction.

        Returns:
            tuple: A tuple containing the minimum and maximum z-values of the layer level.
        """
        return tuple(sorted([self.zmin, self.zmin + self.thickness]))


class LayerStack(BaseModel):
    """For simulation and 3D rendering. Captures design intent of the chip layers after fabrication.

    Parameters:
        layers: dict of layer_levels.
    """

    layers: dict[str, LayerLevel] = Field(
        default_factory=dict,
        description="dict of layer_levels",
    )

    def model_copy(self) -> LayerStack:
        """Returns a copy of the LayerStack."""
        return LayerStack.model_validate_json(self.model_dump_json())

    def __init__(self, **data: Any) -> None:
        """Add LayerLevels automatically for subclassed LayerStacks."""
        super().__init__(**data)

        for field in self.model_dump():
            val = getattr(self, field)
            if isinstance(val, LayerLevel):
                self.layers[field] = val

    def get_layer_to_thickness(self) -> dict[tuple[int, int], float]:
        """Returns layer tuple to thickness (um)."""
        layer_to_thickness = {}

        for level in self.layers.values():
            layer = level.layer

            if layer and level.thickness:
                layer_to_thickness[layer] = level.thickness
            elif hasattr(level, "operator"):
                layer_to_thickness[level.layer] = level.thickness

        return layer_to_thickness

    def get_component_with_derived_layers(self, component, **kwargs):
        """Returns component with derived layers."""

        return get_component_with_derived_layers(
            component=component, layer_stack=self, **kwargs
        )

    def get_layer_to_zmin(self) -> dict[tuple[int, int], float]:
        """Returns layer tuple to z min position (um)."""
        return {
            level.layer: level.zmin for level in self.layers.values() if level.thickness
        }

    def get_layer_to_material(self) -> dict[tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.material
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_sidewall_angle(self) -> dict[tuple[int, int], str]:
        """Returns layer tuple to material name."""
        return {
            level.layer: level.sidewall_angle
            for level in self.layers.values()
            if level.thickness
        }

    def get_layer_to_info(self) -> dict[tuple[int, int], dict]:
        """Returns layer tuple to info dict."""
        return {level.layer: level.info for level in self.layers.values()}

    def get_layer_to_layername(self) -> dict[tuple[int, int], str]:
        """Returns layer tuple to layername."""

        d = defaultdict(list)
        for level_name, level in self.layers.items():
            d[level.layer].append(level_name)

        return d

    def to_dict(self) -> dict[str, dict[str, Any]]:
        return {level_name: dict(level) for level_name, level in self.layers.items()}

    def __getitem__(self, key) -> LayerLevel:
        """Access layer stack elements."""
        if key not in self.layers:
            layers = list(self.layers.keys())
            raise ValueError(f"{key!r} not in {layers}")

        return self.layers[key]

    def get_klayout_3d_script(
        self,
        layer_views: LayerViews | None = None,
        dbu: float | None = 0.001,
    ) -> str:
        """Returns script for 2.5D view in KLayout.

        You can include this information in your tech.lyt

        Args:
            layer_views: optional layer_views.
            dbu: Optional database unit. Defaults to 1nm.
        """
        layers = self.layers or {}

        unetched_layers = [
            layer_name
            for layer_name, level in layers.items()
            if level.layer and level.layer_type == "grow"
        ]
        etch_layers = [
            layer_name
            for layer_name, level in layers.items()
            if level.layer and level.layer_type == "etch"
        ]

        # remove all etched layers from the grown layers
        unetched_layers_dict = defaultdict(list)
        for layer_name in etch_layers:
            level = layers[layer_name]
            into = level.into or []
            for layer_name_etched in into:
                unetched_layers_dict[layer_name_etched].append(layer_name)
                if layer_name_etched in unetched_layers:
                    unetched_layers.remove(layer_name_etched)

        # define layers
        out = "\n".join(
            [
                f"{layer_name} = input({level.layer[0]}, {level.layer[1]})"
                for layer_name, level in layers.items()
                if level.layer
            ]
        )
        out += "\n"
        out += "\n"

        # define unetched layers
        for layer_name_etched, etching_layers in unetched_layers_dict.items():
            etching_layers = " - ".join(etching_layers)
            out += f"unetched_{layer_name_etched} = {layer_name_etched} - {etching_layers}\n"

        out += "\n"

        # define slabs
        for layer_name, level in layers.items():
            if level.layer_type == "etch":
                into = level.into or []
                for i, layer1 in enumerate(into):
                    out += f"slab_{layer1}_{layer_name}_{i} = {layer1} &amp; {layer_name}\n"

        out += "\n"

        for layer_name, level in layers.items():
            layer = level.layer
            zmin = level.zmin
            zmax = zmin + level.thickness
            if dbu:
                rnd_pl = len(str(dbu).split(".")[-1])
                zmin = round(zmin, rnd_pl)
                zmax = round(zmax, rnd_pl)

            if layer is None:
                continue

            elif level.layer_type == "etch":
                name = f"{layer_name}: {level.material}"

                into = level.into or []
                for i, layer1 in enumerate(into):
                    unetched_level = layers[layer1]
                    unetched_zmin = unetched_level.zmin
                    unetched_zmax = unetched_zmin + unetched_level.thickness

                    # slab
                    slab_layer_name = f"slab_{layer1}_{layer_name}_{i}"
                    slab_zmin = unetched_level.zmin
                    slab_zmax = unetched_zmax - level.thickness
                    name = f"{slab_layer_name}: {level.material} {layer[0]}/{layer[1]}"
                    txt = (
                        f"z("
                        f"{slab_layer_name}, "
                        f"zstart: {slab_zmin}, "
                        f"zstop: {slab_zmax}, "
                        f"name: '{name}'"
                    )
                    if layer_views:
                        txt += ", "
                        props = layer_views.get_from_tuple(layer)
                        if hasattr(props, "color"):
                            if props.color.fill == props.color.frame:
                                txt += f"color: {props.color.fill}"
                            else:
                                txt += (
                                    f"fill: {props.color.fill}, "
                                    f"frame: {props.color.frame}"
                                )
                    txt += ")"
                    out += f"{txt}\n"

            elif layer_name in unetched_layers:
                name = f"{layer_name}: {level.material} {layer[0]}/{layer[1]}"

                txt = (
                    f"z("
                    f"{layer_name}, "
                    f"zstart: {zmin}, "
                    f"zstop: {zmax}, "
                    f"name: '{name}'"
                )
                if layer_views:
                    txt += ", "
                    props = layer_views.get_from_tuple(layer)
                    if hasattr(props, "color"):
                        if props.color.fill == props.color.frame:
                            txt += f"color: {props.color.fill}"
                        else:
                            txt += (
                                f"fill: {props.color.fill}, "
                                f"frame: {props.color.frame}"
                            )

                txt += ")"
                out += f"{txt}\n"

        out += "\n"

        for layer_name in unetched_layers_dict:
            unetched_level = self.layers[layer_name]
            layer = unetched_level.layer

            unetched_zmin = unetched_level.zmin
            unetched_zmax = unetched_zmin + unetched_level.thickness
            name = f"{slab_layer_name}: {unetched_level.material}"

            unetched_layer_name = f"unetched_{layer_name}"
            name = f"{unetched_layer_name}: {unetched_level.material} {layer[0]}/{layer[1]}"
            txt = (
                f"z("
                f"{unetched_layer_name}, "
                f"zstart: {unetched_zmin}, "
                f"zstop: {unetched_zmax}, "
                f"name: '{name}'"
            )
            if layer_views:
                txt += ", "
                props = layer_views.get_from_tuple(layer)
                if hasattr(props, "color"):
                    if props.color.fill == props.color.frame:
                        txt += f"color: {props.color.fill}"
                    else:
                        txt += (
                            f"fill: {props.color.fill}, " f"frame: {props.color.frame}"
                        )
            txt += ")"
            out += f"{txt}\n"

        return out

    def filtered(self, layers) -> LayerStack:
        """Returns filtered layerstack, given layer specs."""
        return LayerStack(
            layers={k: self.layers[k] for k in layers if k in self.layers}
        )

    def z_offset(self, dz) -> LayerStack:
        """Translates the z-coordinates of the layerstack."""
        layers = self.layers or {}
        for layer in layers.values():
            layer.zmin += dz

        return self

    def invert_zaxis(self) -> LayerStack:
        """Flips the zmin values about the origin."""
        layers = self.layers or {}
        for layer in layers.values():
            layer.zmin *= -1

        return self


@cell
def get_component_with_derived_layers(component, layer_stack: LayerStack) -> Component:
    """Returns a component with derived layers.

    Args:
        component: Component to get derived layers for.
        layer_stack: Layer stack to get derived layers from.
    """
    unetched_layers = [
        layer_name
        for layer_name, level in layer_stack.layers.items()
        if level.layer and level.layer_type == "grow"
    ]
    etch_layers = [
        layer_name
        for layer_name, level in layer_stack.layers.items()
        if level.layer and level.layer_type == "etch"
    ]

    # remove all etched layers from the grown layers
    unetched_layers_dict = defaultdict(list)
    for layer_name in etch_layers:
        level = layer_stack.layers[layer_name]
        into = level.into or []
        for layer_name_etched in into:
            unetched_layers_dict[layer_name_etched].append(layer_name)
            if layer_name_etched in unetched_layers:
                unetched_layers.remove(layer_name_etched)

    component_layers = component.get_layers()

    # Define pure grown layers
    unetched_layer_numbers = [
        layer_stack.layers[layer_name].layer
        for layer_name in unetched_layers
        if layer_stack.layers[layer_name].layer in component_layers
    ]
    component_derived = component.extract(unetched_layer_numbers)

    # Define unetched layers
    polygons_to_remove = []
    for unetched_layer_name, unetched_layers in unetched_layers_dict.items():
        layer = layer_stack.layers[unetched_layer_name].layer
        polygons = component.get_polygons(by_spec=layer)

        # Add all the etching layers (OR)
        for etching_layers in unetched_layers:
            layer = layer_stack.layers[etching_layers].layer
            B_polys = component.get_polygons(by_spec=layer)
            polygons_to_remove = gdstk.boolean(
                operand1=polygons_to_remove,
                operand2=B_polys,
                operation="or",
                layer=layer[0],
                datatype=layer[1],
            )

            derived_layer = layer_stack.layers[etching_layers].derived_layer
            if derived_layer:
                slab_polygons = gdstk.boolean(
                    operand1=polygons,
                    operand2=B_polys,
                    operation="and",
                    layer=derived_layer[0],
                    datatype=derived_layer[1],
                )
                component_derived.add(slab_polygons)

        # Remove all etching layers
        layer = layer_stack.layers[unetched_layer_name].layer
        polygons = component.get_polygons(by_spec=layer)
        unetched_polys = gdstk.boolean(
            operand1=polygons,
            operand2=polygons_to_remove,
            operation="not",
            layer=layer[0],
            datatype=layer[1],
        )
        component_derived.add(unetched_polys)

    component_derived.add_ports(component.ports)
    return component_derived


if __name__ == "__main__":
    from gdsfactory.generic_tech import get_generic_pdk

    PDK = get_generic_pdk()
    PDK.activate()

    from gdsfactory.generic_tech import LAYER_STACK

    layer_stack = LAYER_STACK

    c = gf.components.straight_heater_metal()
    c.show()

    # import gdsfactory as gf
    # from gdsfactory.generic_tech import LAYER_STACK
    # component = c = gf.components.grating_coupler_elliptical_trenches()
    # component = c = gf.components.taper_strip_to_ridge_trenches()
    # script = LAYER_STACK.get_klayout_3d_script()
    # print(script)
    # ls = layer_stack = LAYER_STACK
    # layer_to_thickness = layer_stack.get_layer_to_thickness()
    # c = layer_stack.get_component_with_derived_layers(component)
    # c.show(show_ports=True)
    # import pathlib
    # filepath = pathlib.Path(
    #     "/home/jmatres/gdslib/sp/temp/write_sparameters_meep_mpi.json"
    # )
    # ls_json = filepath.read_bytes()
    # ls2 = LayerStack.parse_raw(ls_json)
    # from gdsfactory.generic_tech import LAYER_STACK
    # from gdsfactory.technology.klayout_tech import KLayoutTechnology

    # lyp = LayerViews.from_lyp(str(PATH.klayout_lyp))

    # # str_xml = open(PATH.klayout_tech / "tech.lyt").read()
    # # new_tech = db.Technology.technology_from_xml(str_xml)
    # # generic_tech = KLayoutTechnology(layer_views=lyp)
    # connectivity = [("M1", "VIA1", "M2"), ("M2", "VIA2", "M3")]

    # c = generic_tech = KLayoutTechnology(
    #     name="generic_tech", layer_views=lyp, connectivity=connectivity
    # )
    # tech_dir = PATH.klayout_tech
    # # tech_dir = pathlib.Path("/home/jmatres/.klayout/salt/gdsfactory/tech/")
    # tech_dir.mkdir(exist_ok=True, parents=True)
    # generic_tech.write_tech(tech_dir=tech_dir, layer_stack=LAYER_STACK)

    # yaml_test()



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.coupler import coupler
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.straight_pin import straight_pin
from gdsfactory.routing.get_route import get_route
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def mzm(
    phase_shifter: ComponentSpec = straight_pin,
    length_x: float = 500,
    length_y: float = 2.0,
    delta_length: float = 0.0,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight_function,
    splitter: ComponentSpec = "mmi1x2",
    combiner: ComponentSpec | None = "mmi1x2",
    with_splitter: bool = True,
    port_e1_splitter: str = "o2",
    port_e0_splitter: str = "o3",
    port_e1_combiner: str = "o2",
    port_e0_combiner: str = "o3",
    nbends: int = 2,
    cross_section: CrossSectionSpec = "xs_sc",
    mirror_bot: bool = False,
) -> Component:
    """Mzm modulator.

    Args:
        phase_shifter: for bottom and top arms.
        length_x: horizontal length. None uses to the straight_x_bot/top defaults.
        length_y: vertical length for both and top arms.
        delta_length: bottom arm vertical extra length.
        bend: 90 degrees bend spec.
        straight: straight function for vertical.
        splitter: splitter function.
        combiner: combiner function. Optional adds ports.
        with_splitter: if False removes splitter.
        port_e1_splitter: east top splitter port.
        port_e0_splitter: east bot splitter port.
        port_e1_combiner: east top combiner port.
        port_e0_combiner: east bot combiner port.
        nbends: from straight top/bot to combiner (at least 2).
        cross_section: for routing (sxtop/sxbot to combiner).
        mirror_bot: mirrors bottom arm.

    .. code::

                       b2______b3
                      |  sxtop  |
              straight_y        |
                      |         |
                      b1        b4
            splitter==|         |==combiner
                      b5        b8
                      |         |
              straight_y        |
                      |         |
        delta_length/2          |
                      |         |
                     b6__sxbot__b7
                          Lx
    """

    bend_spec = bend
    bend = gf.get_component(bend, cross_section=cross_section)

    c = Component()
    cp1 = gf.get_component(splitter)
    cp2 = gf.get_component(combiner) if combiner else None

    if with_splitter:
        cp1 = c << cp1

    b5 = c << bend
    b5.mirror()
    b5.connect("o1", cp1.ports[port_e0_splitter])

    syl = c << straight(length=delta_length / 2 + length_y, cross_section=cross_section)
    syl.connect("o1", b5.ports["o2"])
    b6 = c << bend
    b6.connect("o1", syl.ports["o2"])

    straight_x_bot = gf.get_component(phase_shifter, length=length_x)
    sxb = c << straight_x_bot
    if mirror_bot:
        sxb.mirror()
    sxb.connect("o1", b6.ports["o2"])

    b1 = c << bend
    b1.connect("o1", cp1.ports[port_e1_splitter])

    sytl = c << straight(length=length_y, cross_section=cross_section)
    sytl.connect("o1", b1.ports["o2"])

    b2 = c << bend
    b2.connect("o2", sytl.ports["o2"])
    straight_x_top = gf.get_component(phase_shifter, length=length_x)
    sxt = c << straight_x_top
    sxt.connect("o1", b2.ports["o1"])

    if combiner:
        cp2 = c << cp2
        cp2.mirror()
        xs = gf.get_cross_section(cross_section)
        cp2.xmin = sxt.ports["o2"].x + bend.info["radius"] * nbends + 2 * xs.min_length

        route = get_route(
            sxt.ports["o2"],
            cp2.ports[port_e1_combiner],
            straight=straight,
            bend=bend_spec,
            cross_section=cross_section,
            with_sbend=False,
        )
        c.add(route.references)
        route = get_route(
            sxb.ports["o2"],
            cp2.ports[port_e0_combiner],
            straight=straight,
            bend=bend_spec,
            cross_section=cross_section,
        )
        c.add(route.references)
        cp2.name = "cp2"

        sytl.name = "sytl"
        syl.name = "syl"
        sxt.name = "sxt"
        sxb.name = "sxb"
        cp1.name = "cp1"
        c.add_ports(cp2.get_ports_list(orientation=0), prefix="out_")
    else:
        c.add_port("o3", port=sxt["o2"])
        c.add_port("o4", port=sxb["o2"])

    if with_splitter:
        c.add_ports(cp1.get_ports_list(orientation=180), prefix="in_")
    else:
        c.add_port("o1", port=b1.ports["o1"])
        c.add_port("o2", port=b5.ports["o1"])

    c.add_ports(sxt.get_ports_list(port_type="electrical"), prefix="top_")
    c.add_ports(sxb.get_ports_list(port_type="electrical"), prefix="bot_")
    c.add_ports(sxt.get_ports_list(port_type="placement"), prefix="top_")
    c.add_ports(sxb.get_ports_list(port_type="placement"), prefix="bot_")
    return c


mzm1x2 = partial(mzm, splitter=mmi1x2, combiner=mmi1x2)
mzm2x2_2x2 = partial(
    mzm,
    splitter=mmi2x2,
    combiner=mmi2x2,
    port_e1_splitter="o3",
    port_e0_splitter="o4",
    port_e1_combiner="o3",
    port_e0_combiner="o4",
)

mzm1x2_2x2 = partial(
    mzm,
    combiner=mmi2x2,
    port_e1_combiner="o3",
    port_e0_combiner="o4",
)

mzm_coupler = partial(
    mzm2x2_2x2,
    splitter=coupler,
    combiner=coupler,
)


if __name__ == "__main__":
    # c = mzm(cache=False)
    c = mzm1x2_2x2(cache=False, delta_length=100)
    c.show(show_ports=True)

    # c = gf.components.mzi2x2_2x2(straight_x_top="straight_heater_metal")
    # c2 = gf.routing.add_fiber_array(c)
    # c2.show()

    # c2 = gf.read.import_gds("a.gds")
    # c3 = gf.grid([c2, c1])
    # c3.show(show_ports=False)



from collections.abc import Iterable

try:
    import kfactory as kf
    from kfactory import KCell, LayerEnum, kcl, kdb
    from kfactory.conf import logger
except ImportError as e:
    print(
        "You can install `pip install gdsfactory[kfactory]` for using maskprep. "
        "And make sure you use python >= 3.10"
    )
    raise e


class FillOperator(kdb.TileOutputReceiver):
    def __init__(
        self,
        kcl: kcl,
        top_cell: KCell,
        fill_cell_index: int,
        fc_bbox: kdb.Box,
        row_step: kdb.Vector,
        column_step: kdb.Vector,
        fill_margin: kdb.Vector = kdb.Vector(0, 0),
        remaining_polygons: kdb.Region | None = None,
    ) -> None:
        self.kcl = kcl
        self.top_cell = top_cell
        self.fill_cell_index = fill_cell_index
        self.fc_bbox = fc_bbox
        self.row_step = row_step
        self.column_step = column_step
        self.fill_margin = fill_margin
        self.remaining_polygons = remaining_polygons
        self.glue_box = self.top_cell.bbox()

    def put(
        self,
        ix: int,
        iy: int,
        tile: kdb.Box,
        region: kdb.Region,
        dbu: float,
        clip: bool,
    ) -> None:
        while not region.is_empty():
            self.top_cell.fill_region(
                region,
                self.fill_cell_index,
                self.fc_bbox,
                self.row_step,
                self.column_step,
                tile.p1,
                region,
                self.fill_margin,
                None,
                self.glue_box,
            )


def fill_tiled(
    c: KCell,
    fill_cell: KCell,
    fill_layers: Iterable[tuple[LayerEnum | int, int]] = [],
    fill_regions: Iterable[tuple[kdb.Region, int]] = [],
    exclude_layers: Iterable[tuple[LayerEnum | int, int]] = [],
    exclude_regions: Iterable[tuple[kdb.Region, int]] = [],
    n_threads: int = 4,
    tile_size: tuple[float, float] | None = None,
    x_space: float = 0,
    y_space: float = 0,
) -> None:
    tp = kdb.TilingProcessor()
    tp.frame = c.bbox().to_dtype(c.kcl.dbu)  # type: ignore
    tp.dbu = c.kcl.dbu
    tp.threads = n_threads

    if tile_size is None:
        tile_size = (
            100 * (fill_cell.dbbox().width() + x_space),
            100 * (fill_cell.dbbox().height() + y_space),
        )
    tp.tile_size(*tile_size)

    layer_names: list[str] = []
    for layer, _ in fill_layers:
        layer_name = f"layer{layer}"
        tp.input(layer_name, c.kcl, c.cell_index(), c.kcl.get_info(layer))
        layer_names.append(layer_name)

    region_names: list[str] = []
    for i, (r, _) in enumerate(fill_regions):
        region_name = f"region{i}"
        tp.input(region_name, r)
        region_names.append(region_name)

    exlayer_names: list[str] = []
    for layer, _ in exclude_layers:
        layer_name = f"layer{layer}"
        tp.input(layer_name, c.kcl, c.cell_index(), c.kcl.get_info(layer))
        exlayer_names.append(layer_name)

    exregion_names: list[str] = []
    for i, (r, _) in enumerate(exclude_regions):
        region_name = f"region{i}"
        tp.input(region_name, r)
        exregion_names.append(region_name)

    tp.output(
        "to_fill",
        FillOperator(
            c.kcl,
            c,
            fill_cell.cell_index(),
            fc_bbox=fill_cell.bbox(),
            row_step=kdb.Vector(fill_cell.bbox().width() + int(x_space / c.kcl.dbu), 0),
            column_step=kdb.Vector(
                0, fill_cell.bbox().height() + int(y_space / c.kcl.dbu)
            ),
        ),
    )

    if layer_names or region_names:
        exlayers = " + ".join(
            [
                layer_name + f".sized({int(size / c.kcl.dbu)})" if size else layer_name
                for layer_name, (_, size) in zip(exlayer_names, exclude_layers)
            ]
        )
        exregions = " + ".join(
            [
                region_name + f".sized({int(size / c.kcl.dbu)})"
                if size
                else region_name
                for region_name, (_, size) in zip(exregion_names, exclude_regions)
            ]
        )
        layers = " + ".join(
            [
                layer_name + f".sized({int(size / c.kcl.dbu)})" if size else layer_name
                for layer_name, (_, size) in zip(layer_names, fill_layers)
            ]
        )
        regions = " + ".join(
            [
                region_name + f".sized({int(size / c.kcl.dbu)})"
                if size
                else region_name
                for region_name, (_, size) in zip(region_names, fill_regions)
            ]
        )

        if exlayer_names or exregion_names:
            queue_str = (
                "var fill= "
                + (
                    " + ".join([layers, regions])
                    if regions and layers
                    else regions + layers
                )
                + "; var exclude = "
                + (
                    " + ".join([exlayers, exregions])
                    if exregions and exlayers
                    else exregions + exlayers
                )
                + "; var fill_region = _tile & _frame & fill - exclude; _output(to_fill, fill_region)"
            )
        else:
            queue_str = (
                "var fill= "
                + (
                    " + ".join([layers, regions])
                    if regions and layers
                    else regions + layers
                )
                + "; var fill_region = _tile & _frame & fill; _output(to_fill, fill_region)"
            )
        tp.queue(queue_str)
        c.kcl.start_changes()
        try:
            logger.info("filling {} with {}", c.name, fill_cell.name)
            logger.debug("fill string: '{}'", queue_str)
            tp.execute(f"Fill {c.name}")
            logger.info("done with filling {}", c.name)
        finally:
            c.kcl.end_changes()


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.geometry.fill_tiled as fill

    c = kf.KCell("ToFill")
    c.shapes(kf.kcl.layer(1, 0)).insert(
        kf.kdb.DPolygon.ellipse(kf.kdb.DBox(5000, 3000), 512)
    )
    c.shapes(kf.kcl.layer(10, 0)).insert(
        kf.kdb.DPolygon(
            [kf.kdb.DPoint(0, 0), kf.kdb.DPoint(5000, 0), kf.kdb.DPoint(5000, 3000)]
        )
    )

    fc = kf.KCell("fill")
    fc.shapes(fc.kcl.layer(2, 0)).insert(kf.kdb.DBox(20, 40))
    fc.shapes(fc.kcl.layer(3, 0)).insert(kf.kdb.DBox(30, 15))

    # fill.fill_tiled(c, fc, [(kf.kcl.layer(1,0), 0)], exclude_layers = [(kf.kcl.layer(10,0), 100), (kf.kcl.layer(2,0), 0), (kf.kcl.layer(3,0),0)], x_space=5, y_space=5)
    fill.fill_tiled(
        c,
        fc,
        [(kf.kcl.layer(1, 0), 0)],
        exclude_layers=[
            (kf.kcl.layer(10, 0), 100),
            (kf.kcl.layer(2, 0), 0),
            (kf.kcl.layer(3, 0), 0),
        ],
        x_space=5,
        y_space=5,
    )

    gdspath = "mzi_fill.gds"
    c.write(gdspath)
    gf.show(gdspath)



from __future__ import annotations

import gdstk

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import Layer


def _union_polygons(polygons, precision: float = 1e-4) -> gdstk.Polygon:
    """Performs union of polygons within PolygonSet or list of polygons.

    Args:
        polygons: PolygonSet or list of polygons. A set containing the input polygons.
        precision: Desired precision for rounding vertex coordinates.
        max_points: maximum number of vertices within the resulting polygon.

    Returns:
        unioned: polygon result of the union of all polygons within the input PolygonSet.
    """
    return gdstk.boolean(
        polygons,
        [],
        operation="or",
        precision=precision,
    )


@gf.cell
def union(
    component: Component,
    by_layer: bool = False,
    precision: float = 1e-4,
    join_first: bool = True,
    layer: Layer = (1, 0),
) -> Component:
    """Returns new Component with inverted union of Component polygons.

    based on phidl.geometry.union

    Args:
        component: Component(/Reference), list of Component(/Reference), or Polygon \
                A containing the polygons to perform union and inversion on.
        by_Layer: performs the union operation layer-wise so each layer can be \
                individually combined.
        precision: Desired precision for rounding vertex coordinates.
        join_first: before offsetting to avoid unnecessary joins in adjacent polygons.
        layer: Specific layer to put polygon geometry on.

    .. plot::
      :include-source:

      import gdsfactory as gf
      c = gf.Component()
      c << gf.components.ellipse(radii=(6, 6))
      c << gf.components.ellipse(radii=(10, 4))
      c2 = gf.geometry.union(c, join_first=False)
      c2.plot()
    """
    U = Component()

    if by_layer:
        all_polygons = component.get_polygons(by_spec=True)
        for layer, polygons in all_polygons.items():
            unioned_polygons = _union_polygons(
                polygons,
                precision=precision,
            )
            U.add_polygon(unioned_polygons, layer=layer)
    else:
        all_polygons = component.get_polygons(by_spec=False)
        unioned_polygons = (
            _union_polygons(
                all_polygons,
                precision=precision,
            )
            if join_first
            else all_polygons
        )
        U.add_polygon(unioned_polygons, layer=layer)
    return U


def test_union() -> None:
    c = Component()
    c << gf.components.ellipse(radii=(6, 6))
    c << gf.components.ellipse(radii=(10, 4))
    c2 = union(c)
    assert int(c2.area()) == 153, c2.area()


if __name__ == "__main__":
    test_union()
    c = Component()
    c << gf.components.ellipse(radii=(6, 6))
    c << gf.components.ellipse(radii=(10, 4))
    c2 = union(c, join_first=False)
    c2.show()



"""Component is a canvas for geometry.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""
from __future__ import annotations

import datetime
import hashlib
import itertools
import math
import os
import pathlib
import uuid
import warnings
from collections import Counter
from collections.abc import Callable, Iterable, Mapping
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import gdstk
import numpy as np
import yaml
from omegaconf import DictConfig

from gdsfactory import snap
from gdsfactory.component_layout import (
    Label,
    _align,
    _distribute,
    _GeometryHelper,
    _parse_layer,
    get_polygons,
    pprint_ports,
)
from gdsfactory.component_reference import ComponentReference, SizeInfo
from gdsfactory.config import CONF, GDSDIR_TEMP, logger
from gdsfactory.name import clean_name, get_name_short
from gdsfactory.polygon import Polygon
from gdsfactory.port import (
    Port,
    auto_rename_ports,
    auto_rename_ports_counter_clockwise,
    auto_rename_ports_layer_orientation,
    auto_rename_ports_orientation,
    map_ports_layer_to_orientation,
    map_ports_to_orientation_ccw,
    map_ports_to_orientation_cw,
    select_ports,
)
from gdsfactory.serialization import clean_dict
from gdsfactory.snap import snap_to_grid

if TYPE_CHECKING:
    from gdsfactory.technology import LayerStack, LayerViews
    from gdsfactory.typings import (
        Coordinate,
        CrossSection,
        CrossSectionSpec,
        Float2,
        Layer,
        Layers,
        LayerSpec,
        PathType,
    )

valid_plotters = ["matplotlib", "klayout", "kweb"]
Axis = Literal["x", "y"]
os.environ["KWEB_FILESLOCATION"] = str(GDSDIR_TEMP)


class UncachedComponentWarning(UserWarning):
    pass


class UncachedComponentError(ValueError):
    pass


class MutabilityError(ValueError):
    pass


def _get_dependencies(component, references_set) -> None:
    for ref in component.references:
        references_set.add(ref.ref_cell)
        _get_dependencies(ref.ref_cell, references_set)


mutability_error_message = """
You cannot modify a Component after creation as it will affect all of its instances.

Create a new Component and add a reference to it.

For example:

# BAD
c = gf.components.bend_euler()
c.add_ref(gf.components.mzi())

# GOOD
c = gf.Component()
c.add_ref(gf.components.bend_euler())
c.add_ref(gf.components.mzi())
"""

move_error_message = """
You cannot move a Component. You can create a new Component, add a reference to the other Component and then move the reference.

For example:

# BAD
c = gf.components.straight()
c.xmin = 10

# GOOD
c = gf.Component()
ref = c.add_ref(gf.components.straight()) # or ref = c << gf.components.straight()
ref.xmin = 10
"""

_timestamp2019 = datetime.datetime.fromtimestamp(1572014192.8273)

# Global dictionary to hold counters for each name
name_counters = Counter()


def _rnd(arr, precision=1e-4):
    arr = np.ascontiguousarray(arr)
    ndigits = round(-math.log10(precision))
    return np.ascontiguousarray(arr.round(ndigits) / precision, dtype=np.int64)


class Component(_GeometryHelper):
    """A Component is an empty canvas where you add polygons, references and ports \
            (to connect to other components).

    - stores settings that you use to build the component
    - stores info that you want to use
    - can return ports by type (optical, electrical ...)
    - can return netlist for circuit simulation
    - can write to GDS, OASIS
    - can show in KLayout, matplotlib, 3D
    - can return copy, mirror, flattened (no references)

    Args:
        name: component_name. Use @cell decorator for auto-naming.
        with_uuid: adds unique identifier.

    Properties:
        info: dictionary that includes
            - derived properties
            - external metadata (test_protocol, docs, ...)
            - simulation_settings
            - function_name
            - name: for the component

        settings:
            full: full settings passed to the function to create component.
            changed: changed settings.
            default: default component settings.
            child: dict info from the children, if any.
    """

    def __init__(
        self,
        name: str = "Unnamed",
        with_uuid: bool = False,
    ) -> None:
        """Initialize the Component object."""

        self.uid = str(uuid.uuid4())[:8]
        if with_uuid or name == "Unnamed":
            name += f"_{self.uid}"

        self._cell = gdstk.Cell("Unnamed")
        self.name = name
        self.info: dict[str, Any] = {}

        self.settings: dict[str, Any] = {}
        self._locked = False
        self._get_child_name = False
        self._reference_names_counter = Counter()
        self._reference_names_used = set()
        self._named_references = {}
        self._references = []

        self.ports = {}

    @property
    def references(self):
        return self._references

    @property
    def polygons(self) -> list[Polygon]:
        return self._cell.polygons

    @property
    def area(self) -> float:
        return self._cell.area

    @property
    def labels(self) -> list[Label]:
        return self._cell.labels

    @property
    def paths(self):
        return self._cell.paths

    @property
    def name(self) -> str:
        return self._cell.name

    @name.setter
    def name(self, name) -> None:
        name = clean_name(name)
        if len(name) > CONF.max_name_length:
            name_short = get_name_short(name)
            warnings.warn(
                f" {name} is too long. Max length is {CONF.max_name_length}. Renaming to {name_short}",
                stacklevel=2,
            )
            name = name_short

        if self.name != name:
            name_counters[name] += 1
            if name_counters[name] > 1:
                name = f"{name}${name_counters[name]-1}"

        self._cell.name = name

    def __iter__(self):
        """You can iterate over polygons, paths, labels and references."""
        return itertools.chain(self.polygons, self.paths, self.labels, self.references)

    def get_polygon_enclosure(self):
        """Returns shapely Polygon with enclosure."""
        import shapely

        return shapely.Polygon(self._cell.convex_hull())

    def get_polygon_bbox(
        self,
        default: float = 0.0,
        top: float | None = None,
        bottom: float | None = None,
        right: float | None = None,
        left: float | None = None,
    ):
        """Returns shapely Polygon with bounding box.

        Args:
            default: default padding in um.
            top: north padding in um.
            bottom: south padding in um.
            right: east padding in um.
            left: west padding in um.
        """
        import shapely

        (xmin, ymin), (xmax, ymax) = self.bbox
        top = top if top is not None else default
        bottom = bottom if bottom is not None else default
        right = right if right is not None else default
        left = left if left is not None else default
        points = [
            [xmin - left, ymin - bottom],
            [xmax + right, ymin - bottom],
            [xmax + right, ymax + top],
            [xmin - left, ymax + top],
        ]
        return shapely.Polygon(points)

    def get_polygons(
        self,
        by_spec: bool | tuple[int, int] = False,
        depth: int | None = None,
        include_paths: bool = True,
        as_array: bool = True,
        as_shapely: bool = False,
        as_shapely_merged: bool = False,
    ) -> list[Polygon] | dict[tuple[int, int], list[Polygon]]:
        """Return a list of polygons in this cell.

        Args:
            by_spec: bool or layer
                If True, the return value is a dictionary with the
                polygons of each individual pair (layer, datatype), which
                are used as keys.  If set to a tuple of (layer, datatype),
                only polygons with that specification are returned.
            depth: integer or None
                If not None, defines from how many reference levels to
                retrieve polygons.  References below this level will result
                in a bounding box.  If `by_spec` is True the key will be the
                name of this cell.
            include_paths: If True, polygonal representation of paths are also included in the result.
            as_array: when as_array=false, return the Polygon objects instead.
                polygon objects have more information (especially when by_spec=False) and are faster to retrieve.
            as_shapely: returns shapely polygons.
            as_shapely_merged: returns a shapely polygonize.

        Returns
            out: list of array-like[N][2] or dictionary
                List containing the coordinates of the vertices of each
                polygon, or dictionary with with the list of polygons (if
                `by_spec` is True).

        Note:
            Instances of `FlexPath` and `RobustPath` are also included in
            the result by computing their polygonal boundary.
        """
        return get_polygons(
            instance=self,
            by_spec=by_spec,
            depth=depth,
            include_paths=include_paths,
            as_array=as_array,
            as_shapely=as_shapely,
            as_shapely_merged=as_shapely_merged,
        )

    def get_dependencies(self, recursive: bool = False) -> list[Component]:
        """Return a list of Components referenced by this Component.

        Args:
            recursive: If True returns dependencies recursively.

        """
        if not recursive:
            return list({ref.parent for ref in self.references})

        references_set = set()
        _get_dependencies(self, references_set=references_set)
        return list(references_set)

    def get_component_spec(self):
        return (
            {
                "component": self.settings.function_name,
                "settings": self.settings.changed,
            }
            if self.settings
            else {"component": self.name, "settings": {}}
        )

    def __getitem__(self, key) -> Port:
        """Access reference ports."""
        if key not in self.ports:
            ports = list(self.ports.keys())
            raise ValueError(f"{key!r} not in {ports}")

        return self.ports[key]

    def __lshift__(self, element) -> ComponentReference:
        """Convenience operator equivalent to add_ref()."""
        return self.add_ref(element)

    def unlock(self) -> None:
        """Only do this if you know what you are doing."""
        self._locked = False

    def lock(self) -> None:
        """Makes sure components can't add new elements or move existing ones.

        Components lock automatically when going into the CACHE to
        ensure one component does not change others
        """
        self._locked = True

    def __setitem__(self, key, element):
        """Allow adding polygons and cell references.

        like D['arc3'] = pg.arc()

        Args:
            key: Alias name.
            element: Object that will be accessible by alias name.

        """
        if isinstance(element, ComponentReference | Polygon):
            self.named_references[key] = element
        else:
            raise ValueError(
                f"Tried to assign alias {key!r} in Component {self.name!r},"
                "but failed because the item was not a ComponentReference"
            )

    @classmethod
    def __get_validators__(cls):
        """Get validators for the Component object."""
        yield cls.validate

    @classmethod
    def validate(cls, v, _info):
        """Pydantic assumes component is valid if the following are true.

        - name characters < pdk.cell_decorator_settings.max_name_length
        - is not empty (has references or polygons)
        """
        from gdsfactory.pdk import get_active_pdk

        pdk = get_active_pdk()

        max_name_length = pdk.cell_decorator_settings.max_name_length
        assert isinstance(
            v, Component
        ), f"TypeError, Got {type(v)}, expecting Component"
        assert (
            len(v.name) <= max_name_length
        ), f"name `{v.name}` {len(v.name)} > {max_name_length} "
        return v

    @property
    def named_references(self):
        return self._named_references

    def add_label(
        self,
        text: str = "hello",
        position: tuple[float, float] = (0.0, 0.0),
        magnification: float = 1.0,
        rotation: float = 0,
        anchor: str = "o",
        layer: LayerSpec = "TEXT",
        x_reflection: bool = False,
    ) -> Label:
        """Adds Label to the Component.

        Args:
            text: Label text.
            position: x-, y-coordinates of the Label location.
            magnification: Magnification factor for the Label text.
            rotation: Angle rotation of the Label text.
            anchor: {'n', 'e', 's', 'w', 'o', 'ne', 'nw', ...}
                Position of the anchor relative to the text.
            layer: Specific layer(s) to put Label on.
            x_reflection: True reflects across the horizontal axis before rotation.
        """
        from gdsfactory.pdk import get_layer

        layer = get_layer(layer)

        gds_layer, gds_datatype = layer

        if not isinstance(text, str):
            text = text
        label = Label(
            text=text,
            origin=position,
            anchor=anchor,
            magnification=magnification,
            rotation=rotation,
            layer=gds_layer,
            texttype=gds_datatype,
            x_reflection=x_reflection,
        )
        self.add(label)
        return label

    @property
    def bbox(self):
        """Returns the bounding box of the ComponentReference."""
        bbox = self._cell.bounding_box()
        if bbox is None:
            bbox = ((0, 0), (0, 0))
        return snap_to_grid(np.array(bbox))

    @property
    def ports_layer(self) -> dict[str, str]:
        """Returns a mapping from layer0_layer1_E0: portName."""
        return map_ports_layer_to_orientation(self.ports)

    def port_by_orientation_cw(self, key: str, **kwargs):
        """Returns port by indexing them clockwise."""
        m = map_ports_to_orientation_cw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def port_by_orientation_ccw(self, key: str, **kwargs):
        """Returns port by indexing them clockwise."""
        m = map_ports_to_orientation_ccw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def get_ports_xsize(self, **kwargs) -> float:
        """Returns xdistance from east to west ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: with in port name.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
        """
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return ports_ccw[0].x - ports_cw[0].x

    def get_ports_ysize(self, **kwargs) -> float:
        """Returns ydistance from east to west ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: with in port name.
            orientation: in degrees.
            width: port width (um).
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
        """
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return ports_ccw[0].y - ports_cw[0].y

    def plot_netlist(
        self, with_labels: bool = True, font_weight: str = "normal", **kwargs
    ):
        """Plots a netlist graph with networkx.

        Args:
            with_labels: add label to each node.
            font_weight: normal, bold.
            **kwargs: keyword arguments for the get_netlist function
        """
        import matplotlib.pyplot as plt
        import networkx as nx

        plt.figure()
        netlist = self.get_netlist(**kwargs)
        connections = netlist["connections"]
        placements = netlist["placements"]
        G = nx.Graph()
        G.add_edges_from(
            [
                (",".join(k.split(",")[:-1]), ",".join(v.split(",")[:-1]))
                for k, v in connections.items()
            ]
        )
        pos = {k: (v["x"], v["y"]) for k, v in placements.items()}
        labels = {k: ",".join(k.split(",")[:1]) for k in placements.keys()}
        nx.draw(
            G,
            with_labels=with_labels,
            font_weight=font_weight,
            labels=labels,
            pos=pos,
        )
        return G

    def plot_netlist_flat(
        self, with_labels: bool = True, font_weight: str = "normal", **kwargs
    ):
        """Plots a netlist graph with networkx.

        Args:
            flat: if true, will plot the flat netlist
            with_labels: add label to each node.
            font_weight: normal, bold.
            **kwargs: keyword arguments for the get_netlist function
        """
        import matplotlib.pyplot as plt
        import networkx as nx

        plt.figure()
        netlist = self.get_netlist_flat(**kwargs)
        connections = netlist["connections"]
        placements = netlist["placements"]
        connections_list = []
        for k, v_list in connections.items():
            connections_list.extend(
                (",".join(k.split(",")[:-1]), ",".join(v.split(",")[:-1]))
                for v in v_list
            )
        G = nx.Graph()
        G.add_edges_from(connections_list)
        pos = {k: (v["x"], v["y"]) for k, v in placements.items()}
        labels = {k: ",".join(k.split(",")[:1]) for k in placements.keys()}
        nx.draw(
            G,
            with_labels=with_labels,
            font_weight=font_weight,
            labels=labels,
            pos=pos,
        )
        return G

    def get_netlist_yaml(self, **kwargs) -> dict[str, Any]:
        from gdsfactory.get_netlist import get_netlist_yaml

        return get_netlist_yaml(self, **kwargs)

    def write_netlist(self, filepath: str) -> None:
        """Write netlist in YAML."""
        netlist = self.get_netlist()
        netlist = clean_dict(netlist)
        filepath = pathlib.Path(filepath)
        filepath.write_text(yaml.dump(netlist))

    def write_netlist_dot(self, filepath: str | None = None) -> None:
        """Write netlist graph in DOT format."""
        from networkx.drawing.nx_agraph import write_dot

        filepath = filepath or f"{self.name}.dot"

        G = self.plot_netlist()
        write_dot(G, filepath)

    def get_netlist(self, **kwargs) -> dict[str, Any]:
        """From Component returns instances, connections and placements dict.

        Keyword Args:
            component: to extract netlist.
            full_settings: True returns all, false changed settings.
            tolerance: tolerance in nm to consider two ports connected.
            exclude_port_types: optional list of port types to exclude from netlisting.
            get_instance_name: function to get instance name.
            allow_multiple: False to raise an error if more than two ports share the same connection.
                if True, will return key: [value] pairs with [value] a list of all connected instances.

        Returns:
            Netlist dict (instances, connections, placements, ports)
                instances: Dict of instance name and settings.
                connections: Dict of Instance1Name,portName: Instance2Name,portName.
                placements: Dict of instance names and placements (x, y, rotation).
                ports: Dict portName: ComponentName,port.
                name: name of component.
        """
        from gdsfactory.get_netlist import get_netlist

        return get_netlist(component=self, **kwargs)

    def get_netlist_recursive(self, **kwargs) -> dict[str, DictConfig]:
        """Returns recursive netlist for a component and subcomponents.

        Keyword Args:
            component: to extract netlist.
            component_suffix: suffix to append to each component name.
                useful if to save and reload a back-annotated netlist.
            get_netlist_func: function to extract individual netlists.
            full_settings: True returns all, false changed settings.
            tolerance: tolerance in nm to consider two ports connected.
            exclude_port_types: optional list of port types to exclude from netlisting.
            get_instance_name: function to get instance name.
            allow_multiple: False to raise an error if more than two ports share the same connection.
                if True, will return key: [value] pairs with [value] a list of all connected instances.

        Returns:
            Dictionary of netlists, keyed by the name of each component.
        """
        from gdsfactory.get_netlist import get_netlist_recursive

        return get_netlist_recursive(component=self, **kwargs)

    def get_netlist_flat(self, **kwargs) -> dict[str, DictConfig]:
        """Returns a netlist where all subinstances are exposed and independently named.

        Keyword Args:
            component: to extract netlist.
            component_suffix: suffix to append to each component name.
                useful if to save and reload a back-annotated netlist.
            get_netlist_func: function to extract individual netlists.
            full_settings: True returns all, false changed settings.
            tolerance: tolerance in nm to consider two ports connected.
            exclude_port_types: optional list of port types to exclude from netlisting.
            get_instance_name: function to get instance name.
            allow_multiple: False to raise an error if more than two ports share the same connection.
                if True, will return key: [value] pairs with [value] a list of all connected instances.

        Returns:
            Dictionary of netlists, keyed by the name of each component.
        """
        from gdsfactory.get_netlist_flat import get_netlist_flat

        return get_netlist_flat(component=self, **kwargs)

    def assert_ports_on_grid(
        self, grid_factor: int = 1, error_type: str = "error"
    ) -> None:
        """Asserts that all ports are on grid."""
        for port in self.ports.values():
            port.assert_on_grid(grid_factor=grid_factor, error_type=error_type)

    def assert_ports_manhattan(self, error_type: str = "error") -> None:
        """Asserts that all ports are on manhattan angles (0, 90, 180, 270)."""
        for port in self.ports.values():
            port.assert_manhattan(error_type=error_type)

    def get_ports(self, depth: int | None = 0):
        """Returns copies of all the ports of the Component, rotated and \
                translated so that they're in their top-level position.

        The Ports returned are copies of the originals, but each copy has the same
        ``uid`` as the original so that they can be traced back to the original if needed.

        Args:
            depth: If not None, defines from how many reference levels to retrieve Ports from.

        Returns:
            port_list : list of Port List of all Ports in the Component.
        """
        port_list = [p._copy() for p in self.ports.values()]

        if depth is None or depth > 0:
            for r in self.references:
                new_depth = None if depth is None else depth - 1
                ref_ports = r.parent.get_ports(depth=new_depth)

                # Transform ports that came from a reference
                ref_ports_transformed = []
                for rp in ref_ports:
                    new_port = rp._copy()
                    new_center, new_orientation = r._transform_port(
                        rp.center,
                        rp.orientation,
                        r.origin,
                        r.rotation,
                        r.x_reflection,
                    )
                    new_port.center = new_center
                    new_port.new_orientation = new_orientation
                    ref_ports_transformed.append(new_port)
                port_list += ref_ports_transformed

        return port_list

    def get_ports_dict(self, **kwargs) -> dict[str, Port]:
        """Returns a dict of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: select ports with prefix in port name.
            suffix: select ports with port name suffix.
            orientation: select ports with orientation in degrees.
            width: select ports with port width.
            layers_excluded: List of layers to exclude.
            port_type: select ports with port_type (optical, electrical, vertical_te).
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return select_ports(self.ports, **kwargs)

    def get_ports_list(self, **kwargs) -> list[Port]:
        """Returns list of ports.

        Keyword Args:
            layer: select ports with GDS layer.
            prefix: select ports with prefix in port name.
            suffix: select ports with port name suffix.
            orientation: select ports with orientation in degrees.
            orientation: select ports with orientation in degrees.
            width: select ports with port width.
            layers_excluded: List of layers to exclude.
            port_type: select ports with port_type (optical, electrical, vertical_te).
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return list(select_ports(self.ports, **kwargs).values())

    def get_ports_pandas(self):
        import pandas as pd

        col_spec = [
            "name",
            "width",
            "center",
            "orientation",
            "layer",
            "port_type",
            "shear_angle",
        ]

        return pd.DataFrame(
            [port.to_dict() for port in self.get_ports_list()], columns=col_spec
        )

    def get_ports_polars(self):
        import polars as pl

        col_spec = {
            "name": pl.Utf8,
            "width": pl.Float64,
            "center": pl.List(pl.Float64),
            "orientation": pl.Float64,
            "layer": pl.List(pl.UInt16),
            "port_type": pl.Utf8,
            "shear_angle": pl.Float64,
        }

        return pl.DataFrame(
            [port.to_dict() for port in self.get_ports_list()], schema=col_spec
        )

    def ref(
        self,
        position: Coordinate = (0, 0),
        port_id: str | None = None,
        rotation: float = 0,
        h_mirror: bool = False,
        v_mirror: bool = False,
    ) -> ComponentReference:
        """Returns Component reference.

        Args:
            position: x, y position.
            port_id: name of the port.
            rotation: in degrees.
            h_mirror: horizontal mirror using y axis (x, 1) (1, 0).
                This is the most common mirror.
            v_mirror: vertical mirror using x axis (1, y) (0, y).
        """
        _ref = ComponentReference(self)

        if port_id and port_id not in self.ports:
            raise ValueError(f"port {port_id} not in {self.ports.keys()}")

        origin = self.ports[port_id].center if port_id else (0, 0)
        if h_mirror:
            _ref.mirror_x(port_id)

        if v_mirror:
            _ref.mirror_y(port_id)

        if rotation != 0:
            _ref.rotate(rotation, origin)
        _ref.move(origin, position)

        return _ref

    def ref_center(self, position=(0, 0)):
        """Returns a reference of the component centered at (x=0, y=0)."""
        si = self.size_info
        yc = si.south + si.height / 2
        xc = si.west + si.width / 2
        center = (xc, yc)
        _ref = ComponentReference(self)
        _ref.move(center, position)
        return _ref

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return (
            f"{self.name}: uid {self.uid}, "
            f"ports {list(self.ports.keys())}, "
            f"references {list(self.named_references.keys())}, "
            f"{len(self.polygons)} polygons"
        )

    def pprint(self) -> None:
        """Prints component info."""
        try:
            from rich import pretty

            pretty.install()
            pretty.pprint(self.to_dict())
        except ImportError:
            print(yaml.dump(self.to_dict()))

    def pprint_ports(self) -> None:
        """Prints ports in a rich table."""

        pprint_ports(self.ports)

    @property
    def metadata_child(self) -> dict:
        """Returns metadata from child if any, Otherwise returns component own.

        metadata Great to access the children metadata at the bottom of the
        hierarchy.
        """
        settings = dict(self.settings)

        while settings.get("child"):
            settings = settings.get("child")

        return dict(settings)

    @property
    def metadata(self) -> dict:
        return dict(self.settings)

    def add_port(
        self,
        name: str | object | None = None,
        center: tuple[float, float] | None = None,
        width: float | None = None,
        orientation: float | None = None,
        port: Port | None = None,
        layer: LayerSpec | None = None,
        port_type: str | None = None,
        cross_section: CrossSectionSpec | None = None,
        shear_angle: float | None = None,
    ) -> Port:
        """Add port to component.

        You can copy an existing port like add_port(port = existing_port) or
        create a new port add_port(myname, mycenter, mywidth, myorientation).
        You can also copy an existing port
        with a new name add_port(port = existing_port, name = new_name)

        Args:
            name: port name.
            center: x, y.
            width: in um.
            orientation: in deg.
            port: optional port.
            layer: port layer.
            port_type: optical, electrical, vertical_dc, vertical_te, vertical_tm. Defaults to optical.
            cross_section: port cross_section.
            shear_angle: an optional angle to shear port face in degrees.
        """
        from gdsfactory.pdk import get_cross_section, get_layer

        layer = get_layer(layer)

        if port:
            if not isinstance(port, Port):
                raise ValueError(f"add_port() needs a Port, got {type(port)}")
            p = port.copy()
            if name is not None:
                p.name = name
            if center is not None:
                p.center = center
            if width is not None:
                p.width = width
            if orientation is not None:
                p.orientation = orientation
            if port_type is not None:
                p.port_type = port_type
            if layer is not None:
                p.layer = layer
            if shear_angle is not None:
                p.shear_angle = shear_angle
            p.parent = self

        elif isinstance(name, Port):
            p = name.copy()
            p.parent = self
            name = p.name
        elif center is None:
            raise ValueError("Port needs center parameter (x, y) um.")

        else:
            p = Port(
                name=name,
                center=center,
                width=width,
                orientation=orientation,
                parent=self,
                layer=layer,
                port_type=port_type or "optical",
                cross_section=get_cross_section(cross_section)
                if cross_section
                else None,
                shear_angle=shear_angle,
            )
            p.parent = self
        if name is not None:
            p.name = name
        if p.name in self.ports:
            raise ValueError(f"add_port() Port name {p.name!r} exists in {self.name!r}")

        self.ports[p.name] = p
        return p

    def add_ports(
        self,
        ports: Iterable[Port] | dict[str, Port],
        prefix: str = "",
        suffix: str = "",
        **kwargs,
    ) -> None:
        """Add a list or dict of ports.

        you can include a prefix to add to the new port names to avoid name conflicts.

        Args:
            ports: list or dict of ports.
            prefix: to prepend to each port name.
            suffix: to append to each port name.
        """
        ports = ports.values() if isinstance(ports, Mapping) else ports
        for port in ports:
            name = f"{prefix}{port.name}{suffix}"
            self.add_port(name=name, port=port, **kwargs)

    def snap_ports_to_grid(self, grid_factor: int = 1) -> None:
        for port in self.ports.values():
            port.snap_to_grid(grid_factor=grid_factor)

    def remove_layers(
        self,
        layers: list[LayerSpec],
        include_labels: bool = True,
        invert_selection: bool = False,
        recursive: bool = True,
    ) -> Component:
        """Remove a list of layers and returns the same Component.

        Args:
            layers: list of layers to remove.
            include_labels: remove labels on those layers.
            invert_selection: removes all layers except layers specified.
            recursive: operate on the cells included in this cell.
        """
        from gdsfactory import get_layer

        component = self.flatten() if recursive and self.references else self
        layers = [get_layer(layer) for layer in layers]

        should_remove = not invert_selection
        component._cell.filter(
            spec=layers,
            remove=should_remove,
            polygons=True,
            paths=True,
            labels=include_labels,
        )
        return component

    def extract(
        self,
        layers: list[tuple[int, int] | str],
    ) -> Component:
        """Extract polygons from a Component and returns a new Component."""
        from gdsfactory.pdk import get_layer

        if type(layers) not in (list, tuple):
            raise ValueError(f"layers {layers!r} needs to be a list or tuple")

        layers = [get_layer(layer) for layer in layers]
        # component = self.copy()
        # component._cell.filter(spec=layers, remove=False)

        component = Component()
        poly_dict = self.get_polygons(by_spec=True, include_paths=False)

        for layer in layers:
            if layer in poly_dict:
                polygons = poly_dict[layer]
                for polygon in polygons:
                    component.add_polygon(polygon, layer)

        for layer in layers:
            for path in self._cell.get_paths(layer=layer):
                component.add(path)

        return component

    def add_polygon(
        self,
        points,
        layer: str | int | tuple[int, int] | np.nan = np.nan,
        snap_to_grid: bool = True,
    ) -> Polygon:
        """Adds a Polygon to the Component.

        Args:
            points: Coordinates of the vertices of the Polygon.
            layer: layer spec to add polygon on.
        """
        from gdsfactory.pdk import get_layer

        if layer is None:
            return None
        elif isinstance(layer, set):
            polygons = [self.add_polygon(points, ly) for ly in layer]
            return polygons[0]

        layer = get_layer(layer)
        if isinstance(points, gdstk.Polygon):
            # if layer is unspecified or matches original polygon, just add it as-is
            polygon = points
            if layer is np.nan or (
                isinstance(layer, tuple) and (polygon.layer, polygon.datatype) == layer
            ):
                polygon = Polygon(polygon.points, (polygon.layer, polygon.datatype))
            else:
                layer, datatype = _parse_layer(layer)
                polygon = Polygon(polygon.points, (layer, datatype))

            if hasattr(points, "properties"):
                polygon.properties = deepcopy(points.properties)

            self._add_polygons(polygon)
            return polygon

        elif hasattr(points, "geoms"):
            for geom in points.geoms:
                polygon = self.add_polygon(geom, layer=layer)
            return polygon
        elif hasattr(points, "exterior"):  # points is a shapely Polygon
            return self._add_polygon_shapely(layer, points)

        points = np.asarray(points)
        if points.ndim == 1:
            return [self.add_polygon(poly, layer=layer) for poly in points]
        if layer is np.nan:
            layer = 0

        if points.ndim == 2:
            # add single polygon from points
            if len(points[0]) > 2:
                # Convert to form [[1,2],[3,4],[5,6]]
                points = np.column_stack(points)
            points = snap.snap_to_grid(points) if snap_to_grid else points
            layer, datatype = _parse_layer(layer)
            polygon = Polygon(points, (layer, datatype))
            self._add_polygons(polygon)
            return polygon
        elif points.ndim == 3:
            layer, datatype = _parse_layer(layer)
            polygons = [Polygon(ppoints, (layer, datatype)) for ppoints in points]
            self._add_polygons(*polygons)
            return polygons
        else:
            raise ValueError(f"Unable to add {points.ndim}-dimensional points object")

    def _add_polygon_shapely(self, layer, points):
        layer, datatype = _parse_layer(layer)
        points_on_grid = snap_to_grid(points.exterior.coords)
        polygon = Polygon(points_on_grid, (layer, datatype))

        if points.interiors:
            return self._add_polygon_shapely_with_holes(
                points, layer, datatype, polygon
            )
        self._add_polygons(polygon)
        return polygon

    def _add_polygon_shapely_with_holes(self, points, layer, datatype, polygon):
        from shapely import get_coordinates

        points_on_grid_interior = np.round(get_coordinates(points.interiors), 3)
        polygon_interior = Polygon(points_on_grid_interior, (layer, datatype))
        polygons = gdstk.boolean(
            polygon,
            polygon_interior,
            operation="not",
            layer=layer,
            datatype=datatype,
        )
        for polygon in polygons:
            self._add_polygons(polygon)
        return polygon

    def _add_polygons(self, *polygons: list[Polygon]) -> None:
        self.is_unlocked()
        self._cell.add(*polygons)

    def copy(self) -> Component:
        return copy(self)

    def add_ref_container(self, component: Component) -> ComponentReference:
        """Add reference, ports and copy_child_info."""
        ref = self << component
        self.add_ports(ref.ports)
        self.copy_child_info(component)
        return ref

    def copy_child_info(self, component: Component) -> None:
        """Copy and settings info from child component into parent.

        Parent components can access child cells settings.
        """
        if not isinstance(component, Component | ComponentReference):
            raise ValueError(
                f"{type(component)}" "is not a Component or ComponentReference"
            )

        self._get_child_name = True
        self.child = component
        self.info.update(component.info)
        self.settings.update(component.settings)

    @property
    def size_info(self) -> SizeInfo:
        """Size info of the component."""
        return SizeInfo(self.bbox)

    def get_setting(self, setting: str) -> str | int | float:
        return (
            self.info.get(setting)
            or self.settings.full.get(setting)
            or self.metadata_child.get(setting)
        )

    def is_unlocked(self) -> None:
        """Raises error if Component is locked."""
        if self._locked:
            raise MutabilityError(
                f"Component {self.name!r} cannot be modified as it's already on cache. "
                + mutability_error_message
            )

    def _add(self, element) -> None:
        """Add a new element or list of elements to this Component.

        Args:
            element: Polygon, ComponentReference or iterable
                The element or iterable of elements to be inserted in this cell.

        Raises:
            MutabilityError: if component is locked.
        """
        self.is_unlocked()
        if isinstance(element, ComponentReference):
            self._cell.add(element._reference)
            self._references.append(element)
        else:
            self._cell.add(element)

    def add(self, element) -> None:
        """Add a new element or list of elements to this Component.

        Args:
            element: Polygon, ComponentReference or iterable
                The element or iterable of elements to be inserted in this cell.

        Raises:
            MutabilityError: if component is locked.
        """
        if isinstance(element, ComponentReference):
            self._register_reference(element)
            self._add(element)
        elif isinstance(element, Iterable):
            for subelement in element:
                self.add(subelement)
        else:
            self._add(element)

    def add_array(
        self,
        component: Component,
        columns: int = 2,
        rows: int = 2,
        spacing: tuple[float, float] = (100, 100),
        alias: str | None = None,
    ) -> ComponentReference:
        """Creates a ComponentReference reference to a Component.

        Args:
            component: The referenced component.
            columns: Number of columns in the array.
            rows: Number of rows in the array.
            spacing: array-like[2] of int or float.
                Distance between adjacent columns and adjacent rows.
            alias: str or None. Alias of the referenced Component.

        Returns
            a: ComponentReference containing references to the Component.
        """
        if not isinstance(component, Component):
            raise TypeError("add_array() needs a Component object.")
        ref = ComponentReference(
            component=component,
            columns=int(round(columns)),
            rows=int(round(rows)),
            spacing=spacing,
        )
        ref.name = None
        self._add(ref)
        self._register_reference(reference=ref, alias=alias)
        return ref

    def distribute(
        self,
        elements: str = "all",
        direction: str = "x",
        spacing: float = 100.0,
        separation: bool = True,
        edge: str = "center",
    ) -> Component:
        """Distributes the specified elements in the Component.

        Args:
            elements: array-like of objects or 'all'. Elements to distribute.
            direction: {'x', 'y'} Direction of distribution; either a line in the x-direction or y-direction.
            spacing  int or float. Distance between elements.
            separation: bool. If True, guarantees elements are separated with a fixed spacing
                between; if  False, elements are spaced evenly along a grid.
            edge: {'x', 'xmin', 'xmax', 'y', 'ymin', 'ymax'}
                Which edge to perform the distribution along (unused if separation == True)

        """
        if elements == "all":
            elements = self.polygons + self.references
        _distribute(
            elements=elements,
            direction=direction,
            spacing=spacing,
            separation=separation,
            edge=edge,
        )
        return self

    def align(self, elements="all", alignment: str = "ymax") -> Component:
        """Align elements in the Component.

        Args:
            elements : array-like of objects, or 'all'
                Elements in the Component to align.
            alignment : {'x', 'y', 'xmin', 'xmax', 'ymin', 'ymax'}
                Which edge to align along (e.g. 'ymax' will move the elements such
                that all of their topmost points are aligned).
        """
        if elements == "all":
            elements = self.polygons + self.references
        _align(elements, alignment=alignment)
        return self

    def flatten(self, single_layer: LayerSpec | None = None) -> Component:
        """Returns a flattened copy of the component.

        Flattens the hierarchy of the Component such that there are no longer
        any references to other Components. All polygons and labels from
        underlying references are copied and placed in the top-level Component.
        If single_layer is specified, all polygons are moved to that layer.

        Args:
            single_layer: move all polygons are moved to the specified (optional).
        """
        component_flat = Component()

        _cell = self._cell.copy(name=component_flat.name)
        _cell = _cell.flatten()
        component_flat._cell = _cell
        if single_layer is not None:
            from gdsfactory import get_layer

            layer, datatype = get_layer(single_layer)
            for polygon in _cell.polygons:
                polygon.layer = layer
                polygon.datatype = datatype
            for path in _cell.paths:
                path.set_layers(layer)
                path.set_datatypes(datatype)

        component_flat.info = self.info.copy()
        component_flat.add_ports(self.ports)
        return component_flat

    def flatten_reference(self, ref: ComponentReference) -> None:
        """From existing cell replaces reference with a flatten reference \
        which has the transformations already applied.

        Transformed reference keeps the original name.

        Args:
            ref: the reference to flatten into a new cell.

        """
        from gdsfactory.functions import transformed

        self.remove(ref)
        new_component = transformed(ref, decorator=None)
        self.add_ref(new_component, alias=ref.name)

    def flatten_invalid_refs(
        self,
        grid_size: float | None = None,
        updated_components=None,
        traversed_components=None,
    ) -> Component:
        """Returns new component with flattened references."""
        return flatten_invalid_refs_recursive(
            self,
            grid_size=grid_size,
            updated_components=updated_components,
            traversed_components=traversed_components,
        )

    def add_ref(
        self, component: Component, alias: str | None = None, **kwargs
    ) -> ComponentReference:
        """Add ComponentReference to the current Component.

        Args:
            component: Component.
            alias: named_references.

        Keyword Args:
            columns: Number of columns in the array.
            rows: Number of rows in the array.
            spacing: Distances between adjacent columns and adjacent rows.
            origin: array-like[2] of int or float
                Position where the cell is inserted.
            rotation : int or float
                Angle of rotation of the reference (in `degrees`).
            magnification : int or float
                Magnification factor for the reference.
            x_reflection : bool
                If True, the reference is reflected parallel to the x direction
                before being rotated.
            name : str (optional)
                A name for the reference (if provided).

        """
        if not isinstance(component, Component):
            raise TypeError(f"type = {type(Component)} needs to be a Component.")
        ref = ComponentReference(component, **kwargs)
        self._add(ref)
        self._register_reference(reference=ref, alias=alias)
        return ref

    def _register_reference(
        self, reference: ComponentReference, alias: str | None = None
    ) -> None:
        component = reference.parent
        reference.owner = self

        if alias is None:
            if reference.name is not None:
                alias = reference.name
            else:
                prefix = (
                    component.settings.function_name
                    if hasattr(component, "settings")
                    and hasattr(component.settings, "function_name")
                    else component.name
                )
                self._reference_names_counter.update({prefix: 1})
                alias = f"{prefix}_{self._reference_names_counter[prefix]}"

                while alias in self._named_references:
                    self._reference_names_counter.update({prefix: 1})
                    alias = f"{prefix}_{self._reference_names_counter[prefix]}"

        reference.name = alias
        self._named_references[alias] = reference

    @property
    def layers(self) -> set[tuple[int, int]]:
        """Returns a set of the Layers in the Component."""
        return self.get_layers()

    def get_layers(self) -> set[tuple[int, int]]:
        """Return a set of (layer, datatype).

        .. code ::

            import gdsfactory as gf
            gf.components.straight().get_layers() == {(1, 0), (111, 0)}
        """
        polygons = self._cell.get_polygons(depth=None)
        return {(polygon.layer, polygon.datatype) for polygon in polygons}

    def get_layer_names(self) -> list[tuple[int, int]]:
        """Return layer names used in the design.

        .. code ::

            import gdsfactory as gf
            gf.components.straight().get_names() == ['WG']
        """
        import gdsfactory as gf

        PDK = gf.get_active_pdk()
        LAYERS = PDK.layers
        name_to_layer = dict(LAYERS)
        layer_to_name = {v: k for k, v in name_to_layer.items()}
        layer_names = []

        for layer in self.layers:
            if layer not in layer_to_name:
                warnings.warn(f"{layer} not in LayerMap.", stacklevel=3)
            else:
                layer_names.append(layer_to_name[layer])
        return layer_names

    def _repr_html_(self):
        """Show geometry in KLayout and in matplotlib for Jupyter Notebooks."""
        self.show(show_ports=True)  # show in klayout
        fig = self.plot()
        if fig and hasattr(fig, "_repr_html_"):
            return fig._repr_html_()

    def add_pins_triangle(
        self,
        port_marker_layer: Layer = (1, 10),
        layer_label: Layer = (1, 10),
        make_copy: bool = True,
    ) -> Component:
        """Returns component with triangular pins."""
        from gdsfactory.add_pins import add_pins_triangle

        if make_copy:
            component = self.copy()
            component.name = self.name
        else:
            component = self
            component.unlock()
        add_pins_triangle(
            component=component, layer=port_marker_layer, layer_label=layer_label
        )
        return component

    def plot_klayout(
        self,
        show_ports: bool = True,
        port_marker_layer: Layer = (1, 10),
        show_labels: bool = False,
    ):
        """Returns klayout image.

        If it fails to import klayout defaults to matplotlib.

        Args:
            show_ports: shows component with port markers and labels.
            port_marker_layer: for the ports.
            show_labels: shows labels.
        """

        component = (
            self.add_pins_triangle(port_marker_layer=port_marker_layer)
            if show_ports
            else self
        )

        try:
            from io import BytesIO

            import klayout.db as db  # noqa: F401
            import klayout.lay as lay
            import matplotlib.pyplot as plt

            from gdsfactory.pdk import get_layer_views

            gdspath = component.write_gds(logging=False)
            lyp_path = gdspath.with_suffix(".lyp")

            layer_views = get_layer_views()
            layer_views.to_lyp(filepath=lyp_path)

            layout_view = lay.LayoutView()
            layout_view.load_layout(str(gdspath.absolute()))
            layout_view.max_hier()
            layout_view.load_layer_props(str(lyp_path))

            layout_view.set_config("text-visible", "true" if show_labels else "false")

            pixel_buffer = layout_view.get_pixels_with_options(800, 600)
            png_data = pixel_buffer.to_png_data()

            # Convert PNG data to NumPy array and display with matplotlib
            with BytesIO(png_data) as f:
                img_array = plt.imread(f)

            # Compute the figure dimensions based on the image size and desired DPI
            dpi = 80
            fig_width = img_array.shape[1] / dpi
            fig_height = img_array.shape[0] / dpi

            fig, ax = plt.subplots(figsize=(fig_width, fig_height), dpi=dpi)

            # Remove margins and display the image
            ax.imshow(img_array)
            ax.axis("off")  # Hide axes
            ax.set_position([0, 0, 1, 1])  # Set axes to occupy the full figure space

            plt.subplots_adjust(
                left=0, right=1, top=1, bottom=0, wspace=0, hspace=0
            )  # Remove any padding
            plt.tight_layout(pad=0)  # Ensure no space is wasted
            return fig

        except ImportError:
            print("You can install `pip install gdsfactory` for better visualization")
            component.plot(plotter="matplotlib")

    def plot_kweb(self):
        """Shows current gds in kweb."""

        try:
            import kweb.server_jupyter as kj
        except Exception:
            print("You need to install kweb with `pip install gdsfactory[cad]`")
            return self.plot_klayout()

        from html import escape

        from IPython.display import IFrame

        from gdsfactory.pdk import get_layer_views

        gdspath = self.write_gds(gdsdir=GDSDIR_TEMP, logging=False)
        lyp_path = GDSDIR_TEMP / "layers.lyp"

        layer_props = get_layer_views()
        layer_props.to_lyp(filepath=lyp_path)
        host = os.getenv("KWEB_HOST", "localhost")

        port = (
            kj.port
            if hasattr(kj, "port") and kj.port
            else int(os.getenv("KWEB_PORT", 8000))
        )
        src = f"http://{host}:{port}/gds/{escape(gdspath.stem+gdspath.suffix)}?layer_props={escape(str(lyp_path))}"

        os.environ["KWEB_PORT"] = str(os.getenv("KWEB_PORT", port))

        if not kj.jupyter_server:
            port = port
            while kj.is_port_in_use(port=port, host=host):
                port += 1

            os.environ["KWEB_PORT"] = str(port)
            logger.debug(src)
            kj.start()

        if kj.jupyter_server:
            return IFrame(
                src=src,
                width=1400,
                height=600,
            )
        else:
            return self.plot_klayout()

    def plot_matplotlib(self, **kwargs) -> None:
        """Plot component using matplotlib.

        Keyword Args:
            show_ports: Sets whether ports are drawn.
            show_subports: Sets whether subports (ports that belong to references) are drawn.
            label_aliases: Sets whether aliases are labeled with a text name.
            new_window: If True, each call to quickplot() will generate a separate window.
            blocking: If True, calling quickplot() will pause execution of ("block") the
                remainder of the python code until the quickplot() window is closed.
                If False, the window will be opened and code will continue to run.
            zoom_factor: Sets the scaling factor when zooming the quickplot window with the
                mousewheel/trackpad.
            interactive_zoom: Enables using mousewheel/trackpad to zoom.
            fontsize: for labels.
            layers_excluded: list of layers to exclude.
            layer_views: layer_views colors loaded from Klayout.
            min_aspect: minimum aspect ratio.
        """
        from gdsfactory.quickplotter import quickplot

        quickplot(self, **kwargs)

    def plot(self, plotter: str | None = None, **kwargs):
        """Returns component plot using klayout, matplotlib, or kweb.

        We recommend using klayout or kweb.
        Klayout is good for images and kweb for responsive interactive plots.
        Matplotlib is slow for rendering big layouts and is almost Deprecated.

        Args:
            plotter: plot backends ('widget', 'klayout', 'kweb').
        """
        plotter = plotter or CONF.display_type

        if plotter not in valid_plotters:
            raise ValueError(f"{plotter!r} not in {valid_plotters}")

        if plotter == "klayout":
            self.plot_klayout()
            return
        elif plotter == "kweb":
            return self.plot_kweb()

        elif plotter == "matplotlib":
            from gdsfactory.quickplotter import quickplot

            quickplot(self, **kwargs)
            return

    def show(
        self,
        show_ports: bool = False,
        show_subports: bool = False,
        port_marker_layer: Layer = (1, 10),
        **kwargs,
    ) -> None:
        """Show component in KLayout.

        returns a copy of the Component, so the original component remains intact.
        with pins markers on each port show_ports = True, and optionally also
        the ports from the references (show_subports=True)

        Args:
            show_ports: shows component with port markers and labels.
            show_subports: add ports markers and labels to references.
            port_marker_layer: for the ports.

        Keyword Args:
            gdspath: GDS file path to write to.
            gdsdir: directory for the GDS file. Defaults to /tmp/.
            unit: unit size for objects in library. 1um by default.
            precision: for object dimensions in the library (m). 1nm by default.
            timestamp: Defaults to 2019-10-25. If None uses current time.
        """
        from gdsfactory.add_pins import add_pins_triangle
        from gdsfactory.show import show

        component = (
            self.add_pins_triangle(
                port_marker_layer=port_marker_layer,
                layer_label=port_marker_layer,
                make_copy=False,
            )
            if show_ports
            else self
        )

        if show_subports:
            component = self.copy()
            for reference in component.references:
                if isinstance(component, ComponentReference):
                    add_pins_triangle(
                        component=component,
                        reference=reference,
                        layer=port_marker_layer,
                        layer_label=port_marker_layer,
                    )

        show(component, **kwargs)

    def _write_library(
        self,
        gdspath: PathType | None = None,
        gdsdir: PathType | None = None,
        timestamp: datetime.datetime | None = _timestamp2019,
        logging: bool = True,
        with_oasis: bool = False,
        with_metadata: bool = False,
        with_netlist: bool = False,
        netlist_function: Callable | None = None,
        **kwargs,
    ) -> Path:
        """Write component to GDS or OASIS and returns gdspath.

        Args:
            gdspath: GDS file path to write to.
            gdsdir: directory for the GDS file. Defaults to /tmp/randomFile/gdsfactory.
            timestamp: Defaults to 2019-10-25 for consistent hash.
                If None uses current time.
            logging: disable GDS path logging, for example for showing it in KLayout.
            with_oasis: If True, file will be written to OASIS. Otherwise, file will be written to GDS.
            with_metadata: writes metadata in YAML format.
            with_netlist: writes netlist in JSON format.
            netlist_function: The netlist function to use. You can compose a partial function with the `get_netlist` function for example with your parameters.

        Keyword Args:
            Keyword arguments will override the active PDK's default GdsWriteSettings and OasisWriteSettings.

            Gds settings:
                unit: unit size for objects in library. 1um by default.
                precision: for dimensions in the library (m). 1nm by default.
                on_duplicate_cell: specify how to resolve duplicate-named cells. Choose one of the following:
                    "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                    "error": throw a ValueError when attempting to write a gds with duplicate cells.
                    "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.
                    None: do not try to resolve (at your own risk!)
                flatten_invalid_refs: flattens component references which have invalid transformations.
                max_points: Maximal number of vertices per polygon. Polygons with more vertices that this are automatically fractured.

            Oasis settings:
                compression_level: Level of compression for cells (between 0 and 9).
                    Setting to 0 will disable cell compression, 1 gives the best speed and 9, the best compression.
                detect_rectangles: Store rectangles in compressed format.
                detect_trapezoids: Store trapezoids in compressed format.
                circle_tolerance: Tolerance for detecting circles. If less or equal to 0, no detection is performed. Circles are stored in compressed format.
                validation ("crc32", "checksum32", None): type of validation to include in the saved file.
                standard_properties: Store standard OASIS properties in the file.

        """

        from gdsfactory.pdk import get_active_pdk

        if gdspath and gdsdir:
            warnings.warn(
                "gdspath and gdsdir have both been specified. gdspath will take precedence and gdsdir will be ignored.",
                stacklevel=3,
            )

        default_settings = get_active_pdk().gds_write_settings
        default_oasis_settings = get_active_pdk().oasis_settings

        explicit_gds_settings = {
            k: v
            for k, v in kwargs.items()
            if v is not None and k in default_settings.model_dump()
        }
        explicit_oas_settings = {
            k: v
            for k, v in kwargs.items()
            if v is not None and k in default_oasis_settings.model_dump()
        }
        # update the write settings with any settings explicitly passed
        write_settings = default_settings.model_copy(update=explicit_gds_settings)
        oasis_settings = default_oasis_settings.model_copy(update=explicit_oas_settings)

        _check_uncached_components(
            component=self, mode=write_settings.on_uncached_component
        )

        if write_settings.flatten_invalid_refs:
            top_cell = flatten_invalid_refs_recursive(self)
        else:
            top_cell = self

        gdsdir = gdsdir or GDSDIR_TEMP
        gdsdir = pathlib.Path(gdsdir)
        if with_oasis:
            gdspath = gdspath or gdsdir / f"{top_cell.name}.oas"
        else:
            gdspath = gdspath or gdsdir / f"{top_cell.name}.gds"
        gdspath = pathlib.Path(gdspath)
        gdsdir = gdspath.parent
        gdsdir.mkdir(exist_ok=True, parents=True)

        cells = top_cell.get_dependencies(recursive=True)
        cell_names = [cell.name for cell in list(cells)]
        cell_names_unique = set(cell_names)

        if len(cell_names) != len(set(cell_names)):
            for cell_name in cell_names_unique:
                cell_names.remove(cell_name)

            if write_settings.on_duplicate_cell == "error":
                raise ValueError(
                    f"Duplicated cell names in {top_cell.name!r}: {cell_names!r}"
                )
            elif write_settings.on_duplicate_cell in {"warn", "overwrite"}:
                if write_settings.on_duplicate_cell == "warn":
                    warnings.warn(
                        f"Duplicated cell names in {top_cell.name!r}:  {cell_names}",
                        stacklevel=3,
                    )
                cells_dict = {cell.name: cell._cell for cell in cells}
                cells = cells_dict.values()
            elif write_settings.on_duplicate_cell is not None:
                raise ValueError(
                    f"on_duplicate_cell: {write_settings.on_duplicate_cell!r} not in (None, warn, error, overwrite)"
                )

        all_cells = [top_cell._cell] + sorted(cells, key=lambda cc: cc.name)

        no_name_cells = [
            cell.name for cell in all_cells if cell.name.startswith("Unnamed")
        ]

        if no_name_cells:
            warnings.warn(
                f"Unnamed cells, {len(no_name_cells)} in {top_cell.name!r}",
                stacklevel=3,
            )

        # for cell in all_cells:
        #     print(cell.name, type(cell))

        lib = gdstk.Library(
            name=write_settings.lib_name,
            unit=write_settings.unit,
            precision=write_settings.precision,
        )
        lib.add(top_cell._cell)
        lib.add(*top_cell._cell.dependencies(True))

        if with_oasis:
            lib.write_oas(gdspath, **oasis_settings.dict())
        else:
            lib.write_gds(
                gdspath, timestamp=timestamp, max_points=write_settings.max_points
            )
        if logging:
            logger.info(f"Wrote to {str(gdspath)!r}")
        if with_metadata:
            metadata = gdspath.with_suffix(".yml")
            metadata.write_text(self.to_yaml(with_cells=True, with_ports=True))
            logger.info(f"Write YAML metadata to {str(metadata)!r}")

        if with_netlist:
            """
            Saves the netlist_function output to a json file.
            """
            import json

            if netlist_function is None:
                from gdsfactory.get_netlist import get_netlist

                netlist_function = get_netlist

            netlist_path = gdspath.with_suffix(".json")
            netlist_dictionary = netlist_function(component=self, **kwargs)
            netlist_path.write_text(json.dumps(netlist_dictionary, indent=2))

        CONF.last_saved_files.append(gdspath)
        return gdspath

    def write_gds(
        self,
        gdspath: PathType | None = None,
        gdsdir: PathType | None = None,
        **kwargs,
    ) -> Path:
        """Write component to GDS and returns gdspath.

        Args:
            gdspath: GDS file path to write to.
            gdsdir: directory for the GDS file. Defaults to /tmp/randomFile/gdsfactory.

        Keyword Args:
            unit: unit size for objects in library. 1um by default.
            precision: for dimensions in the library (m). 1nm by default.
            logging: disable GDS path logging, for example for showing it in KLayout.
            on_duplicate_cell: specify how to resolve duplicate-named cells. Choose one of the following:
                "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                "error": throw a ValueError when attempting to write a gds with duplicate cells.
                "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.
            on_uncached_component: Literal["warn", "error"] = "warn"
            flatten_invalid_refs: flattens component references which have invalid transformations.
            max_points: Maximal number of vertices per polygon.
                Polygons with more vertices that this are automatically fractured.
            with_metadata: writes metadata in YAML format.
            with_netlist: writes a netlist in JSON format.
            netlist_function: function to generate the netlist.
        """

        return self._write_library(
            gdspath=gdspath, gdsdir=gdsdir, with_oasis=False, **kwargs
        )

    def write_oas(
        self,
        gdspath: PathType | None = None,
        gdsdir: PathType | None = None,
        **kwargs,
    ) -> Path:
        """Write component to GDS and returns gdspath.

        Args:
            gdspath: GDS file path to write to.
            gdsdir: directory for the GDS file. Defaults to /tmp/randomFile/gdsfactory.

        Keyword Args:
            unit: unit size for objects in library. 1um by default.
            precision: for dimensions in the library (m). 1nm by default.
            logging: disable GDS path logging, for example for showing it in KLayout.
            on_duplicate_cell: specify how to resolve duplicate-named cells. Choose one of the following:
                "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                "error": throw a ValueError when attempting to write a gds with duplicate cells.
                "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.
                None: do not try to resolve (at your own risk!)
            on_uncached_component: Literal["warn", "error"] = "warn"
            flatten_invalid_refs: flattens component references which have invalid transformations.
            compression_level: Level of compression for cells (between 0 and 9).
                Setting to 0 will disable cell compression, 1 gives the best speed and 9, the best compression.
            detect_rectangles: Store rectangles in compressed format.
            detect_trapezoids: Store trapezoids in compressed format.
            circle_tolerance: Tolerance for detecting circles. If less or equal to 0, no detection is performed.
                Circles are stored in compressed format.
            validation ("crc32", "checksum32", None) – type of validation to include in the saved file.
            standard_properties: Store standard OASIS properties in the file.
        """
        return self._write_library(
            gdspath=gdspath,
            gdsdir=gdsdir,
            with_oasis=True,
            **kwargs,
        )

    def to_dict(
        self,
        ignore_components_prefix: list[str] | None = None,
        ignore_functions_prefix: list[str] | None = None,
        with_cells: bool = False,
        with_ports: bool = True,
    ) -> dict[str, Any]:
        """Returns Dict representation of a component.

        Args:
            ignore_components_prefix: for components to ignore when exporting.
            ignore_functions_prefix: for functions to ignore when exporting.
            with_cells: write cells recursively.
            with_ports: write port information dict.
        """
        d = {}
        if with_ports:
            ports = {port.name: port.to_dict() for port in self.get_ports_list()}
            d["ports"] = ports

        if with_cells:
            cells = recurse_structures(
                self,
                ignore_functions_prefix=ignore_functions_prefix,
                ignore_components_prefix=ignore_components_prefix,
            )
            d["cells"] = clean_dict(cells)

        d["name"] = self.name
        d["settings"] = clean_dict(dict(self.settings))
        return d

    def to_yaml(self, **kwargs) -> str:
        """Write Dict representation of a component in YAML format.

        Args:
            ignore_components_prefix: for components to ignore when exporting.
            ignore_functions_prefix: for functions to ignore when exporting.
            with_cells: write cells recursively.
            with_ports: write port information.
        """
        return yaml.dump(clean_dict(self.to_dict(**kwargs)))

    def auto_rename_ports(self, **kwargs) -> None:
        """Rename ports by orientation NSEW (north, south, east, west).

        Keyword Args:
            function: to rename ports.
            select_ports_optical: to select optical ports.
            select_ports_electrical: to select electrical ports.
            prefix_optical: prefix.
            prefix_electrical: prefix.

        .. code::

                  3  4
                 _|__|_
             2 -|      |- 5
                |      |
             1 -|______|- 6
                  |  |
                  8  7
        """
        self.is_unlocked()
        auto_rename_ports(self, **kwargs)

    def auto_rename_ports_counter_clockwise(self, **kwargs) -> None:
        self.is_unlocked()
        auto_rename_ports_counter_clockwise(self, **kwargs)

    def auto_rename_ports_layer_orientation(self, **kwargs) -> None:
        self.is_unlocked()
        auto_rename_ports_layer_orientation(self, **kwargs)

    def auto_rename_ports_orientation(self, **kwargs) -> None:
        """Rename ports by orientation NSEW (north, south, east, west).

        Keyword Args:
            function: to rename ports.
            select_ports_optical: to select ports.
            select_ports_electrical:
            prefix_optical:
            prefix_electrical:

        .. code::

                 N0  N1
                 |___|_
            W1 -|      |- E1
                |      |
            W0 -|______|- E0
                 |   |
                S0   S1
        """
        self.is_unlocked()
        auto_rename_ports_orientation(self, **kwargs)

    def move(self, *args, **kwargs) -> Component:
        """Make a reference instead"""
        raise ValueError(move_error_message)

    def mirror(self, p1: Float2 = (0, 1), p2: Float2 = (0, 0), **kwargs) -> Component:
        """Returns new Component with a mirrored reference.

        Args:
            p1: first point to define mirror axis.
            p2: second point to define mirror axis.
        """
        from gdsfactory.functions import mirror

        return mirror(component=self, p1=p1, p2=p2, **kwargs)

    def rotate(self, angle: float = 90, **kwargs) -> Component:
        """Returns new component with a rotated reference to the original.

        Args:
            angle: in degrees.
        """
        from gdsfactory.functions import rotate

        return rotate(component=self, angle=angle, **kwargs)

    def add_padding(self, **kwargs) -> Component:
        """Returns same component with padding.

        Keyword Args:
            component: for padding.
            layers: list of layers.
            suffix for name.
            default: default padding (50um).
            top: north padding.
            bottom: south padding.
            right: east padding.
            left: west padding.
        """
        from gdsfactory.add_padding import add_padding

        return add_padding(component=self, **kwargs)

    def absorb(self, reference) -> Component:
        """Absorbs polygons from ComponentReference into Component.

        Destroys the reference in the process but keeping the polygon geometry.

        Args:
            reference: ComponentReference to be absorbed into the Component.
        """
        if reference not in self.references:
            raise ValueError(
                "The reference you asked to absorb does not exist in this Component."
            )
        ref_polygons = reference.get_polygons(
            by_spec=False, include_paths=False, as_array=False
        )
        self._add_polygons(*ref_polygons)

        self.add(reference.get_labels())
        self.add(reference.get_paths())
        self.remove(reference)
        return self

    def remove(self, items):
        """Removes items from a Component, which can include Ports, PolygonSets \
        CellReferences, ComponentReferences and Labels.

        Args:
            items: list of Items to be removed from the Component.
        """
        if not hasattr(items, "__iter__"):
            items = [items]
        for item in items:
            if isinstance(item, Port):
                self.ports = {k: v for k, v in self.ports.items() if v != item}
            elif isinstance(item, gdstk.Reference):
                self._cell.remove(item)
                item.owner = None
            elif isinstance(item, ComponentReference):
                self.references.remove(item)
                self._cell.remove(item._reference)
                item.owner = None
                self._named_references.pop(item.name)
            else:
                self._cell.remove(item)

        self._bb_valid = False
        return self

    def hash_geometry(self, precision: float = 1e-4) -> str:
        """Returns an SHA1 hash of the geometry in the Component.

        For each layer, each polygon is individually hashed and then the polygon hashes
        are sorted, to ensure the hash stays constant regardless of the ordering
        the polygons.  Similarly, the layers are sorted by (layer, datatype).

        Args:
            precision: Rounding precision for the the objects in the Component.
                For instance, a precision of 1e-2 will round a point at
                (0.124, 1.748) to (0.12, 1.75).

        """
        polygons_by_spec = self.get_polygons(by_spec=True, as_array=False)
        layers = np.array(list(polygons_by_spec.keys()))
        sorted_layers = layers[np.lexsort((layers[:, 0], layers[:, 1]))]

        final_hash = hashlib.sha1()
        for layer in sorted_layers:
            layer_hash = hashlib.sha1(layer.astype(np.int64)).digest()
            polygons = polygons_by_spec[tuple(layer)]
            polygons = [_rnd(p.points, precision) for p in polygons]
            polygon_hashes = np.sort([hashlib.sha1(p).digest() for p in polygons])
            final_hash.update(layer_hash)
            for ph in polygon_hashes:
                final_hash.update(ph)

        return final_hash.hexdigest()

    def get_labels(
        self, apply_repetitions=True, depth: int | None = None, layer=None
    ) -> list[Label]:
        """Return labels.

        Args:
            apply_repetitions:.
            depth: None returns all labels and 0 top level.
            layer: layerspec.
        """
        from gdsfactory.pdk import get_layer

        if layer:
            layer, texttype = get_layer(layer)
        else:
            texttype = None
        return self._cell.get_labels(
            apply_repetitions=apply_repetitions,
            depth=depth,
            layer=layer,
            texttype=texttype,
        )

    def remove_labels(self) -> None:
        """Remove labels."""
        self._cell.remove(*self.labels)

    # Deprecated
    def get_info(self):
        """Gathers the .info dictionaries from every sub-Component and returns them in a list.

        Args:
            depth: int or None
                If not None, defines from how many reference levels to
                retrieve Ports from.

        Returns:
            list of dictionaries
                List of the ".info" property dictionaries from all sub-Components
        """
        D_list = self.get_dependencies(recursive=True)
        return [D.info.copy() for D in D_list]

    def remap_layers(
        self, layermap, include_labels: bool = True, include_paths: bool = True
    ) -> Component:
        """Returns a copy of the component with remapped layers.

        Args:
            layermap: Dictionary of values in format {layer_from: layer_to}.
            include_labels: Selects whether to move Labels along with polygons.
            include_paths: Selects whether to move Paths along with polygons.
        """
        component = self.copy()
        layermap = {_parse_layer(k): _parse_layer(v) for k, v in layermap.items()}

        all_D = list(component.get_dependencies(True))
        all_D.append(component)
        for D in all_D:
            for p in D.polygons:
                layer = (p.layer, p.datatype)
                if layer in layermap:
                    new_layer = layermap[layer]
                    p.layer = new_layer[0]
                    p.datatype = new_layer[1]
            if include_labels:
                for label in D.labels:
                    original_layer = (label.layer, label.texttype)
                    original_layer = _parse_layer(original_layer)
                    if original_layer in layermap:
                        new_layer = layermap[original_layer]
                        label.layer = new_layer[0]
                        label.texttype = new_layer[1]

            if include_paths:
                for path in D.paths:
                    new_layers = list(path.layers)
                    new_datatypes = list(path.datatypes)
                    for layer_number in range(len(new_layers)):
                        original_layer = _parse_layer(
                            (new_layers[layer_number], new_datatypes[layer_number])
                        )
                        if original_layer in layermap:
                            new_layer = layermap[original_layer]
                            new_layers[layer_number] = new_layer[0]
                            new_datatypes[layer_number] = new_layer[1]
                    path.set_layers(*new_layers)
                    path.set_datatypes(*new_datatypes)
        return component

    def to_3d(
        self,
        layer_views: LayerViews | None = None,
        layer_stack: LayerStack | None = None,
        exclude_layers: tuple[Layer, ...] | None = None,
    ):
        """Return Component 3D trimesh Scene.

        Args:
            component: to extrude in 3D.
            layer_views: layer colors from Klayout Layer Properties file.
                Defaults to active PDK.layer_views.
            layer_stack: contains thickness and zmin for each layer.
                Defaults to active PDK.layer_stack.
            exclude_layers: layers to exclude.

        """
        from gdsfactory.export.to_3d import to_3d

        return to_3d(
            self,
            layer_views=layer_views,
            layer_stack=layer_stack,
            exclude_layers=exclude_layers,
        )

    def to_np(
        self,
        nm_per_pixel: int = 20,
        layers: Layers = ((1, 0),),
        values: tuple[float, ...] | None = None,
        pad_width: int = 1,
    ) -> np.ndarray:
        """Returns a pixelated numpy array from Component polygons.

        Args:
            component: Component.
            nm_per_pixel: you can go from 20 (coarse) to 4 (fine).
            layers: to convert. Order matters (latter overwrite former).
            values: associated to each layer (defaults to 1).
            pad_width: padding pixels around the image.

        """
        from gdsfactory.export.to_np import to_np

        return to_np(
            self,
            nm_per_pixel=nm_per_pixel,
            layers=layers,
            values=values,
            pad_width=pad_width,
        )

    def write_stl(
        self,
        filepath: str,
        layer_stack: LayerStack | None = None,
        exclude_layers: tuple[Layer, ...] | None = None,
        use_layer_name: bool = False,
        hull_invalid_polygons: bool = True,
        scale: float | None = None,
    ) -> None:
        """Write a Component to STL for 3D printing.

        Args:
            filepath: to write STL to.
            layer_stack: contains thickness and zmin for each layer.
            exclude_layers: layers to exclude.
            use_layer_name: If True, uses LayerLevel names in output filenames rather than gds_layer and gds_datatype.
            hull_invalid_polygons: If True, replaces invalid polygons (determined by shapely.Polygon.is_valid) with its convex hull.
            scale: Optional factor by which to scale meshes before writing.

        """
        from gdsfactory.export.to_stl import to_stl

        to_stl(
            self,
            filepath=filepath,
            layer_stack=layer_stack,
            exclude_layers=exclude_layers,
            use_layer_name=use_layer_name,
            hull_invalid_polygons=hull_invalid_polygons,
            scale=scale,
        )

    def write_gerber(self, dirpath, layermap_to_gerber_layer, options) -> None:
        """
        Args:
            dirpath: directory to write gerber files to.
            layermap_to_gerber_layer: dictionary of layermap to gerber layer.
            options: dictionary of options for gerber export.
                header: List[str] | None = None
                mode: Literal["mm", "in"] = "mm"
                resolution: float = 1e-6
                int_size: int = 4
        """
        from gdsfactory.export.to_gerber import to_gerber

        to_gerber(
            self,
            dirpath=dirpath,
            layermap_to_gerber_layer=layermap_to_gerber_layer,
            options=options,
        )

    def to_gmsh(self, *args, **kwargs) -> None:
        """Deprecated. instead of.

        mesh = component.to_gmsh(arguments)

        Use:

        from gplugins.gmsh.get_mesh import get_mesh

        """

        raise ValueError(
            """component.to_gmsh() has been deprecated. Instead of:

        mesh = component.to_gmsh(arguments)

        Use:

        from gplugins.gmsh.get_mesh import get_mesh

        mesh = get_mesh(component, arguments)
        """
        )

    def offset(
        self,
        distance: float = 0.1,
        use_union: bool = True,
        precision: float = 1e-4,
        join: str = "miter",
        tolerance: int = 2,
        layer: LayerSpec = "WG",
    ) -> Component:
        """Returns new Component with polygons eroded or dilated by an offset.

        Args:
        distance: Distance to offset polygons. Positive values expand, negative shrink.
        use_union: If True, use union of all polygons to offset. If False, offset
        precision: Desired precision for rounding vertex coordinates.
        join: {'miter', 'bevel', 'round'} Type of join used to create polygon offset
        tolerance: For miter joints, this number must be at least 2 represents the
          maximal distance in multiples of offset between new vertices and their
          original position before beveling to avoid spikes at acute joints. For
          round joints, it indicates the curvature resolution in number of
          points per full circle.
        layer: Specific layer to put polygon geometry on.

        """
        from gdsfactory.geometry.offset import offset

        return offset(
            self,
            distance=distance,
            use_union=use_union,
            precision=precision,
            join=join,
            tolerance=tolerance,
            layer=layer,
        )

    def add_route_info(
        self,
        cross_section: CrossSection | str,
        length: float,
        length_eff: float | None = None,
        taper: bool = False,
        **kwargs,
    ) -> None:
        """Adds route information to a component.

        Args:
            cross_section: CrossSection or name of the cross_section.
            length: length of the route.
            length_eff: effective length of the route.
            taper: if True adds taper information.
            **kwargs: extra information to add to the component.
        """
        from gdsfactory.pdk import get_active_pdk

        pdk = get_active_pdk()

        length_eff = length_eff or length
        xs_name = (
            cross_section
            if isinstance(cross_section, str)
            else pdk.get_cross_section_name(cross_section)
        )

        d = {
            "type": xs_name,
            "length": length_eff,
            f"{xs_name}_length": length_eff,
            "weight": length_eff,
        }
        if taper:
            d[f"{xs_name}_taper_length"] = length
        d |= kwargs
        self.info["route_info"] = d


# Component methods


def copy(
    D: Component,
    references=None,
    ports=None,
    polygons=None,
    paths=None,
    name=None,
    labels=None,
) -> Component:
    """Returns a Component copy.

    Args:
        D: component to copy.
        references: references to copy.
        ports: ports to copy.
        polygons: polygons to copy.
        paths: paths to copy.
        name: name of the new component.
        labels: labels to copy.
    """
    c = Component()
    c.info = D.info

    for ref in references if references is not None else D.references:
        c.add(copy_reference(ref))
    for port in (ports if ports is not None else D.ports).values():
        c.add_port(port=port)
    for poly in polygons if polygons is not None else D.polygons:
        c.add_polygon(poly)
    for path in paths if paths is not None else D.paths:
        c.add(path)
    for label in labels if labels is not None else D.labels:
        c.add_label(
            text=label.text,
            position=label.origin,
            layer=(label.layer, label.texttype),
        )

    if name is not None:
        c.name = name

    return c


def copy_reference(
    ref,
    parent=None,
    columns=None,
    rows=None,
    spacing=None,
    origin=None,
    rotation=None,
    magnification=None,
    x_reflection=None,
    name=None,
    v1=None,
    v2=None,
) -> ComponentReference:
    return ComponentReference(
        component=parent or ref.parent,
        columns=columns or ref.columns,
        rows=rows or ref.rows,
        spacing=spacing or ref.spacing,
        origin=origin or ref.origin,
        rotation=rotation or ref.rotation,
        magnification=magnification or ref.magnification,
        x_reflection=x_reflection or ref.x_reflection,
        name=name or ref.name,
        v1=v1 or ref.v1,
        v2=v2 or ref.v2,
    )


def _filter_polys(polygons, layers_excl):
    return [
        polygon
        for polygon, layer, datatype in zip(
            polygons.polygons, polygons.layers, polygons.datatypes
        )
        if (layer, datatype) not in layers_excl
    ]


def recurse_structures(
    component: Component,
    ignore_components_prefix: list[str] | None = None,
    ignore_functions_prefix: list[str] | None = None,
) -> dict[str, Any]:
    """Recurse component and components references recursively.

    Args:
        component: component to recurse.
        ignore_components_prefix: list of prefix to ignore.
        ignore_functions_prefix: list of prefix to ignore.
    """
    ignore_functions_prefix = ignore_functions_prefix or []
    ignore_components_prefix = ignore_components_prefix or []

    if (
        hasattr(component, "function_name")
        and component.function_name in ignore_functions_prefix
    ):
        return {}

    if hasattr(component, "name") and any(
        component.name.startswith(i) for i in ignore_components_prefix
    ):
        return {}

    output = {component.name: dict(component.settings)}
    for reference in component.references:
        if (
            isinstance(reference, ComponentReference)
            and reference.ref_cell.name not in output
        ):
            output.update(recurse_structures(reference.ref_cell))

    return output


def flatten_invalid_refs_recursive(
    component: Component,
    grid_size: float | None = None,
    updated_components=None,
    traversed_components=None,
) -> Component:
    """Recursively flattens component references which have invalid transformations
    (i.e. non-90 deg rotations or sub-grid translations)
    returns a copy if any subcells have been modified.

    WARNING: this function will produce same-name copies of cells.
    It is strictly meant to be used on write of the GDS file and
    should not be mixed with other cells,
    or you will likely experience issues with duplicate cells

    Args:
        component: the component to fix (in place).
        grid_size: the GDS grid size, in um, defaults to active PDK.get_grid_size()
            any translations with higher resolution than this are considered invalid.
        updated_components: dict of components transformed.
            Should always be None, except for recursive.
        traversed_components: the set of component names which have been traversed.
            Should always be None, except for recursive invocations.
    """
    from gdsfactory.decorators import is_invalid_ref

    invalid_refs = []
    refs = component.references
    subcell_modified = False
    if updated_components is None:
        updated_components = {}
    if traversed_components is None:
        traversed_components = set()
    for ref in refs:
        # mark any invalid refs for flattening
        # otherwise, check if there are any modified cells beneath (we need not do this if the ref will be flattened anyways)
        if is_invalid_ref(ref, grid_size):
            invalid_refs.append(ref.name)
        else:
            # otherwise, recursively flatten refs if the subcell has not already been traversed
            if ref.parent.name not in traversed_components:
                flatten_invalid_refs_recursive(
                    ref.parent,
                    grid_size=grid_size,
                    updated_components=updated_components,
                    traversed_components=traversed_components,
                )
            # now, if the ref's cell been modified, mark it as such
            if ref.parent.name in updated_components:
                subcell_modified = True
    if invalid_refs or subcell_modified:
        new_component = component.copy()
        new_component.name = f"{component.name}_t"
        # make sure all modified cells have their references updated
        new_refs = new_component.references.copy()
        for ref in new_refs:
            if ref.name in invalid_refs:
                new_component.flatten_reference(ref)
            elif (
                ref.parent.name in updated_components
                and ref.parent is not updated_components[ref.parent.name]
            ):
                ref.parent = updated_components[ref.parent.name]
        component = new_component
        updated_components[component.name] = new_component
    traversed_components.add(component.name)
    return component


def _check_uncached_components(component, mode):
    valid_modes = ["warn", "error", "ignore"]

    if mode == "ignore":
        return
    elif mode not in valid_modes:
        raise ValueError(
            f"{mode} is not a valid value for on_uncached_component. Try one of these: {valid_modes}."
        )

    for sub_component in component.get_dependencies(recursive=True):
        if not sub_component._locked:
            message = (
                f"Component {sub_component.name!r} was NOT properly locked. "
                "You need to write it into a function that has the @cell decorator."
            )
            if mode == "warn":
                warnings.warn(message, UncachedComponentWarning, stacklevel=3)

            elif mode == "error":
                raise UncachedComponentError(message)


if __name__ == "__main__":
    import gdsfactory as gf

    gf.config.enable_off_grid_ports()

    c = gf.Component("bend")
    b = c << gf.components.bend_circular(angle=30)
    s = c << gf.components.straight(length=5)
    s.connect("o1", b.ports["o2"])
    p_shapely = c.get_polygons(as_shapely_merged=True)
    c2 = gf.Component("bend_fixed")
    c2.add_polygon(p_shapely, layer=(1, 0))
    c2.plot()

    c = gf.c.mzi(flatten=True, decorator=gf.routing.add_fiber_single)
    # print(c.name)
    c.show()

    # c = gf.c.mzi()
    # fig = c.plot_klayout()
    # fig.savefig("mzi.png")
    # c.pprint_ports()

    # c = gf.Component("hi" * 200)
    # print(c.name)

    # c = gf.Component("hi" * 200)
    # print(c.name)
    # p = c.add_polygon(
    #     [(-8, 6, 7, 9), (-6, 8, 17, 5)], layer=(1, 0)
    # )  # GDS layers are tuples of ints (but if we use only one number it assumes the other number is 0)
    # c.write_gds("hi.gds")
    # c.show()
    # print(CONF.last_saved_files)



"""When you create components you have to make sure they have unique names.

the cell decorator gives unique names to components that depend on their parameters.
"""

from __future__ import annotations

import gdsfactory as gf

if __name__ == "__main__":
    c1 = gf.components.straight(length=5)
    assert c1.name.split("_")[0] == "straight"



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.straight import straight
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def coupler_asymmetric(
    bend: ComponentSpec = bend_s,
    gap: float = 0.234,
    dy: float = 2.5,
    dx: float = 10.0,
    cross_section: CrossSectionSpec = "xs_sc_no_pins",
) -> Component:
    """Bend coupled to straight waveguide.

    Args:
        bend: spec.
        gap: um.
        dy: port to port vertical spacing.
        dx: bend length in x direction.
        cross_section: spec.

    .. code::

                        dx
                     |-----|
                      _____ o2
                     /         |
               _____/          |
         gap o1____________    |  dy
                            o3
    """
    x = gf.get_cross_section(cross_section)
    width = x.width
    bend_component = (
        bend(size=(dx, dy - gap - width), cross_section=cross_section)
        if callable(bend)
        else bend
    )
    wg = straight(cross_section=cross_section)

    w = bend_component.ports["o1"].width
    y = (w + gap) / 2

    c = Component()
    wg = wg.ref(position=(0, y), port_id="o1")
    bottom_bend = bend_component.ref(position=(0, -y), port_id="o1", v_mirror=True)

    c.add(wg)
    c.add(bottom_bend)
    c.absorb(wg)
    c.absorb(bottom_bend)

    port_width = 2 * w + gap
    c.add_port(
        name="o1", center=(0, 0), width=port_width, orientation=180, cross_section=x
    )
    c.add_port(name="o3", port=bottom_bend.ports["o2"])
    c.add_port(name="o2", port=wg.ports["o2"])
    return c


if __name__ == "__main__":
    c = coupler_asymmetric()
    c.show(show_ports=False)



import sys
import warnings

try:
    import gplugins
except ImportError as e:
    raise ImportError('You need to install gplugins with "pip install gplugins"') from e


message = """
gdsfactory.simulation have been moved to gplugins

Make sure you have gplugins installed and use gplugins instead of gdsfactory.simulation

You can replace:
    import gdsfactory.simulation -> import gplugins

You can install gplugins with:
    pip install gplugins
"""

warnings.warn(message)
sys.modules["gdsfactory.simulation"] = gplugins



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_circular import bend_circular
from gdsfactory.components.straight import straight
from gdsfactory.cross_section import strip
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def ring_crow(
    gaps: list[float] = [0.2] * 4,
    radius: list[float] = [10.0] * 3,
    input_straight_cross_section: CrossSectionSpec = strip,
    output_straight_cross_section: CrossSectionSpec = strip,
    bends: list[ComponentSpec] = [bend_circular] * 3,
    ring_cross_sections: list[CrossSectionSpec] = [strip] * 3,
) -> Component:
    """Coupled ring resonators.

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler: ring coupler spec.
        straight: straight spec.
        bend: bend spec.
        cross_section: cross_section spec.
        kwargs: cross_section settings.

    .. code::

         --==ct==-- gap[N-1]
          |      |
          sl     sr ring[N-1]
          |      |
         --==cb==-- gap[N-2]

             .
             .
             .

         --==ct==--
          |      |
          sl     sr lengths_y[1], ring[1]
          |      |
         --==cb==-- gap[1]

         --==ct==--
          |      |
          sl     sr lengths_y[0], ring[0]
          |      |
         --==cb==-- gap[0]

          length_x
    """
    c = Component()

    # Input bus
    input_straight = gf.get_component(
        straight, length=2 * radius[0], cross_section=input_straight_cross_section
    )
    input_straight_cross_section = gf.get_cross_section(input_straight_cross_section)
    input_straight_width = input_straight_cross_section.width

    input_straight_waveguide = c.add_ref(input_straight).movex(-radius[0])
    c.add_port(name="o1", port=input_straight_waveguide.ports["o1"])
    c.add_port(name="o2", port=input_straight_waveguide.ports["o2"])

    # Cascade rings
    cum_y_dist = input_straight_width / 2

    for index, (gap, r, bend, cross_section) in enumerate(
        zip(gaps, radius, bends, ring_cross_sections)
    ):
        gap = gf.snap.snap_to_grid(gap, grid_factor=2)
        ring = Component(f"ring{index}")

        bend_c = gf.get_component(bend, radius=r, cross_section=cross_section)
        xs = gf.get_cross_section(cross_section)
        bend_width = xs.width
        bend1 = ring.add_ref(bend_c, alias=f"bot_right_bend_ring_{index}")
        bend2 = ring.add_ref(bend_c, alias=f"top_right_bend_ring_{index}")
        bend3 = ring.add_ref(bend_c, alias=f"top_left_bend_ring_{index}")
        bend4 = ring.add_ref(bend_c, alias=f"bot_left_bend_ring_{index}")

        bend2.connect("o1", bend1.ports["o2"])
        bend3.connect("o1", bend2.ports["o2"])
        bend4.connect("o1", bend3.ports["o2"])

        ring_ref = c.add_ref(ring)
        ring_ref.movey(cum_y_dist + gap + bend_width / 2)
        c.absorb(ring_ref)
        cum_y_dist += gap + bend_width + 2 * r

    # Output bus
    output_straight = gf.get_component(
        straight,
        length=2 * radius[-1],
        cross_section=output_straight_cross_section,
    )
    output_straight_width = output_straight_cross_section().width
    output_straight_waveguide = (
        c.add_ref(output_straight)
        .movey(cum_y_dist + gaps[-1] + output_straight_width / 2)
        .movex(-radius[-1])
    )
    c.add_port(name="o3", port=output_straight_waveguide.ports["o1"])
    c.add_port(name="o4", port=output_straight_waveguide.ports["o2"])
    return c


if __name__ == "__main__":
    c = ring_crow(
        input_straight_cross_section="xs_rc",
        ring_cross_sections=["xs_rc", "xs_sc", "xs_rc"],
    )
    # c = ring_crow(gaps = [0.3, 0.4, 0.5, 0.2],
    #     radius = [20.0, 5.0, 15.0],
    #     input_straight_cross_section = strip,
    #     output_straight_cross_section = strip,
    #     bends = [bend_circular] * 3,
    #     ring_cross_sections = [strip] * 3,
    # )

    c.show(show_ports=True, show_subports=False)



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_s import bend_s
from gdsfactory.typings import ComponentFactory, CrossSectionSpec


@gf.cell
def mmi2x2_with_sbend(
    with_sbend: bool = True,
    s_bend: ComponentFactory = bend_s,
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Returns mmi2x2 for Cband.

    C_band 2x2MMI in 220nm thick silicon
    https://opg.optica.org/oe/fulltext.cfm?uri=oe-25-23-28957&id=376719

    Args:
        with_sbend: add sbend.
        s_bend: S-bend function.
        cross_section: spec.
    """

    def mmi_widths(t):
        return np.array([2 * 0.7 + 0.2, 1.48, 1.48, 1.48, 1.6])

    c = gf.Component()

    P = gf.path.straight(length=2 * 2.4 + 2 * 1.6, npoints=5)
    xs = gf.get_cross_section(cross_section)
    xs0 = xs.copy(width_function=mmi_widths, add_pins_function_name=None)
    ref = c << gf.path.extrude(P, cross_section=xs0)

    # Add input and output tapers
    taper = gf.components.taper(
        length=1, width1=0.5, width2=0.7, cross_section=cross_section, add_pins=False
    )
    topl_taper = c << taper
    topl_taper.move((-1, 0.45))
    botl_taper = c << taper
    botl_taper.move((-1, -0.45))

    topr_taper = c << taper
    topr_taper.mirror()
    topr_taper.move((9, 0.45))

    botr_taper = c << taper
    botr_taper.mirror()
    botr_taper.move((9, -0.45))

    if with_sbend:
        sbend = s_bend(cross_section=cross_section, add_pins=False)

        topl_sbend = c << sbend
        topl_sbend.mirror([0, 1])
        botl_sbend = c << sbend
        topr_sbend = c << sbend
        botr_sbend = c << sbend
        botr_sbend.mirror([0, 1])

        topl_sbend.connect("o1", destination=topl_taper.ports["o1"])
        botl_sbend.connect("o1", destination=botl_taper.ports["o1"])
        topr_sbend.connect("o1", destination=topr_taper.ports["o1"])
        botr_sbend.connect("o1", destination=botr_taper.ports["o1"])

        c.add_port("o1", port=botl_sbend.ports["o2"])
        c.add_port("o2", port=topl_sbend.ports["o2"])
        c.add_port("o3", port=topr_sbend.ports["o2"])
        c.add_port("o4", port=botr_sbend.ports["o2"])

        c.absorb(topr_sbend)
        c.absorb(topl_sbend)
        c.absorb(botr_sbend)
        c.absorb(botl_sbend)
    else:
        c.add_port("o2", port=topl_taper.ports["o1"])
        c.add_port("o1", port=botl_taper.ports["o1"])
        c.add_port("o3", port=topr_taper.ports["o1"])
        c.add_port("o4", port=botr_taper.ports["o1"])

    xs.add_pins(c)
    c.absorb(ref)
    c.absorb(topr_taper)
    c.absorb(topl_taper)
    c.absorb(botr_taper)
    c.absorb(botl_taper)
    return c


if __name__ == "__main__":
    # c = mmi2x2_with_sbend(
    #     with_sbend=True,
    #     cross_section=dict(cross_section="xs_sc", settings=dict(layer=(2, 0))),
    # )
    c = mmi2x2_with_sbend()
    c.show(show_ports=True)



"""You can define a function to add pins."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.add_pins import add_pins_triangle

if __name__ == "__main__":
    c = gf.components.straight(decorator=add_pins_triangle)
    c.show(show_ports=False)



"""You can also use the fiber array routing functions for connecting to pads."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.samples.big_device import big_device

if __name__ == "__main__":
    w = h = 18 * 50
    cross_section = "xs_m3"
    c = big_device(nports=10, cross_section=cross_section)
    c = gf.routing.add_fiber_array(
        component=c,
        cross_section=cross_section,
        grating_coupler="pad",
        gc_port_name="e1",
        get_input_labels_function=None,
        with_loopback=False,
        straight_separation=15,
        # bend='wire_corner'
    )
    c.show()



"""Read uPDK YAML definition and returns a gdsfactory script.

https://openepda.org/index.html
"""

from __future__ import annotations

import io
import pathlib
from typing import IO

from omegaconf import OmegaConf

from gdsfactory.typings import LayerSpec, PathType


def from_updk(
    filepath: PathType,
    filepath_out: PathType | None = None,
    layer_bbox: tuple[int, int] = (68, 0),
    layer_bbmetal: tuple[int, int] | None = None,
    layer_label: tuple[int, int] | None = None,
    layer_pin_label: tuple[int, int] | None = None,
    layer_pin: tuple[int, int] | None = None,
    optical_xsections: list[str] | None = None,
    electrical_xsections: list[str] | None = None,
    layers_text: list[LayerSpec] | None = None,
    text_size: float = 2.0,
    activate_pdk: bool = False,
    read_xsections: bool = True,
    prefix: str = "",
    suffix: str = "",
) -> str:
    """Read uPDK definition and returns a gdsfactory script.

    Args:
        filepath: uPDK filepath definition.
        filepath_out: optional filepath to save script. if None only returns script and does not save it.
        layer_bbox: layer to draw bounding boxes.
        layer_bbmetal: layer to draw bounding boxes for metal.
        optical_xsections: Optional list of names of xsections that will add optical ports.
        electrical_xsections: Optional list of names of xsections that will add electrical ports.
        layers_text: Optional list of layers to add text labels.
        text_size: text size for labels.
        activate_pdk: if True, activate the pdk after writing the script.
        read_xsections: if True, read xsections from uPDK.
        prefix: optional prefix to add to the script.
        suffix: optional suffix to add to the script.
    """

    optical_xsections = optical_xsections or []
    electrical_xsections = electrical_xsections or []

    if isinstance(filepath, str | pathlib.Path | IO):
        filepath = (
            io.StringIO(filepath)
            if isinstance(filepath, str) and "\n" in filepath
            else filepath
        )

        conf = OmegaConf.load(
            filepath
        )  # nicer loader than conf = yaml.safe_load(filepath)
    else:
        conf = OmegaConf.create(filepath)

    script = prefix
    script += f"""

import sys
import gdsfactory as gf
from gdsfactory.get_factories import get_cells
from gdsfactory.add_pins import add_pins_inside2um

layer_bbox = {layer_bbox}
layer_bbmetal = {layer_bbmetal}
layer_pin_label = {layer_pin_label}
layer_pin = {layer_pin}
"""

    if layer_label:
        script += f"layer_label = {layer_label}\n"

    if read_xsections and "xsections" in conf:
        for xsection_name, xsection in conf.xsections.items():
            script += f"{xsection_name} = gf.CrossSection(width={xsection.width})\n"

        xs = ",".join([f"{name}={name}" for name in conf.xsections.keys()])
        script += "\n"
        script += f"cross_sections = dict({xs})"
        script += "\n"

    for block_name, block in conf.blocks.items():
        parameters = block.parameters
        parameters_string = (
            ", ".join(
                [f"{p_name}:{p.type}={p.value}" for p_name, p in parameters.items()]
            )
            if parameters
            else ""
        )
        parameters_doc = (
            "\n    ".join(
                [
                    f"  {p_name}: {p.doc} (min: {p.min}, max: {p.max}, {p.unit})."
                    for p_name, p in parameters.items()
                    if hasattr(p, "min")
                ]
            )
            if parameters
            else ""
        )

        parameters_colon = (
            [f"{p_name}:{{{p_name}}}" for p_name in parameters] if parameters else []
        )
        parameters_equal = (
            [f"{p_name}={{{p_name}}}" for p_name in parameters] if parameters else []
        )

        parameters_labels = (
            "\n".join(
                [
                    "    c.add_label(text=f'{}', position=(xc, yc-{}/{}/2*ysize), layer=layer_label)\n".format(
                        p_name, i, len(parameters)
                    )
                    for i, p_name in enumerate(parameters_colon)
                ]
            )
            if layer_label and parameters_colon
            else ""
        )

        if parameters:
            doc = f'"""{block.doc}\n\n    Args:\n    {parameters_doc}\n    """'
        else:
            doc = f'"""{block.doc}"""'

        cell_name = (
            f"{block_name}:{','.join(parameters_equal)}"
            if parameters_equal
            else block_name
        )

        points = str(block.bbox).replace("'", "")
        script += f"""
@gf.cell
def {block_name}({parameters_string})->gf.Component:
    {doc}
    c = gf.Component()
    p = c.add_polygon({points}, layer=layer_bbox)
    xc, yc = p.center
    ysize = p.ysize
    name = f{cell_name!r}
"""
        if layer_bbmetal and "bb_metal" in block:
            for bbmetal in block["bb_metal"].values():
                points = str(bbmetal).replace("'", "")
                script += f"    c.add_polygon({points}, layer=layer_bbmetal)\n"

        script += parameters_labels

        for port_name, port in block.pins.items():
            port_type = (
                "electrical" if port.xsection in electrical_xsections else "optical"
            )

            if port.xsection != "None":
                script += f"    c.add_port(name={port_name!r}, width={port.width}, cross_section={port.xsection!r}, center=({port.xya[0]}, {port.xya[1]}), orientation={port.xya[2]}, port_type={port_type!r})\n"
            else:
                script += f"    c.add_port(name={port_name!r}, width={port.width}, layer=(0, 0), center=({port.xya[0]}, {port.xya[1]}), orientation={port.xya[2]}, port_type={port_type!r})\n"

            if layer_pin_label:
                d = OmegaConf.to_container(port)
                d["name"] = port_name
                text = OmegaConf.to_yaml(d)
                script += f"    c.add_label(text={text!r}, position=({port.xya[0]}, {port.xya[1]}), layer=layer_pin_label)\n"
        if layers_text:
            for layer_text in layers_text:
                script += f"    text = c << gf.c.text(text=name, size={text_size}, position=(xc, yc), layer={layer_text},justify='center')\n"
                script += "    c.absorb(text)\n"

        script += """
    c.name = name
    if layer_pin:
        add_pins_inside2um(c, layer=layer_pin)
    return c
"""

    if activate_pdk:
        script += f"""
cells = get_cells(sys.modules[__name__])
pdk = gf.Pdk(name={conf.header.description!r}, cells=cells, cross_sections=cross_sections)
pdk.activate()
"""

    script += f"""
{suffix}

if __name__ == "__main__":
    c = {block_name}()
    c.show(show_ports=True)
"""
    if filepath_out:
        filepath_out = pathlib.Path(filepath_out)
        filepath_out.write_text(script)
    return script


if __name__ == "__main__":
    from gdsfactory.samples.pdk.fab_c import pdk

    yaml_pdk_decription = pdk.to_updk()
    gdsfactory_script = from_updk(yaml_pdk_decription)



from __future__ import annotations

from functools import partial

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.via_stack import via_stack_slab_npp_m3
from gdsfactory.typings import ComponentSpec, Floats, LayerSpecs

pad_via_stack_slab_npp = partial(via_stack_slab_npp_m3, size=(80, 80))


@cell
def resistance_sheet(
    width: float = 10,
    layers: LayerSpecs = ("SLAB90", "NPP"),
    layer_offsets: Floats = (0, 0.2),
    pad: ComponentSpec = pad_via_stack_slab_npp,
    pad_pitch: float = 100.0,
    ohms_per_square: float | None = None,
    port_orientation1: int = 180,
    port_orientation2: int = 0,
) -> Component:
    """Returns Sheet resistance.

    keeps connectivity for pads and first layer in layers

    Args:
        width: in um.
        layers: for the middle part.
        layer_offsets: from edge, positive: over, negative: inclusion.
        pad: function to create a pad.
        pad_pitch: in um.
        ohms_per_square: optional sheet resistance to compute info.resistance.
        port_orientation1: in degrees.
        port_orientation2: in degrees.
    """
    c = Component()

    pad = pad()
    length = pad_pitch - pad.get_setting("size")[0]

    pad1 = c << pad
    pad2 = c << pad
    r0 = c << compass(
        size=(length + layer_offsets[0], width + layer_offsets[0]), layer=layers[0]
    )
    c.absorb(r0)

    for layer, offset in zip(layers[1:], layer_offsets[1:]):
        r = c << compass(size=(length + 2 * offset, width + 2 * offset), layer=layer)
        c.absorb(r)

    pad1.connect("e3", r0.ports["e1"])
    pad2.connect("e1", r0.ports["e3"])

    c.info["resistance"] = ohms_per_square * width * length if ohms_per_square else None

    c.add_port(
        "pad1",
        port_type="vertical_dc",
        center=pad1.center,
        layer=list(layers)[-1],
        width=width,
        orientation=port_orientation1,
    )
    c.add_port(
        "pad2",
        port_type="vertical_dc",
        center=pad2.center,
        layer=list(layers)[-1],
        width=width,
        orientation=port_orientation2,
    )
    c.absorb(pad1)
    c.absorb(pad2)
    return c


if __name__ == "__main__":
    # import gdsfactory as gf
    # sweep = [resistance_sheet(width=width, layers=((1,0), (1,1))) for width in [1, 10, 100]]
    # c = gf.pack(sweep)[0]

    c = resistance_sheet()
    c.show(show_ports=True)

    # import gdsfactory as gf
    # sweep_resistance = list(map(resistance_sheet, (5, 10, 80)))
    # c = gf.grid(sweep_resistance)
    # c.show(show_ports=True)
